##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Mapping Tests"""
import doctest
import persistent
import pprint
import transaction

from pjpersist import testing, mapping

class Item(persistent.Persistent):
    def __init__(self, name=None, site=None):
        self.name = name
        self.site = site

def doctest_PJTableMapping_simple():
    r"""PJTableMapping: simple

    The PJ Table Mapping provides a Python dict interface for a PostGreSQL
    table. Here is a simple example for our Item class/table:

      >>> class SimpleContainer(mapping.PJTableMapping):
      ...     __pj_table__ = 'pjpersist_dot_tests_dot_test_mapping_dot_Item'
      ...     __pj_mapping_key__ = 'name'

    To initialize the mapping, we need a data manager:

      >>> container = SimpleContainer(dm)

    Let's do some obvious initial manipulations:

      >>> container['one'] = one = Item()
      >>> one.name
      'one'
      >>> transaction.commit()

    After the transaction is committed, we can access the item:

      >>> container.keys()
      ['one']
      >>> container['one'].name
      'one'

      >>> container['two']
      Traceback (most recent call last):
      ...
      KeyError: 'two'

    Of course we can delete an item, but note that it only removes the name,
    but does not delete the document by default:

      >>> del container['one']
      >>> transaction.commit()
      >>> container.keys()
      []
    """

def doctest_PJTableMapping_filter():
    r"""PJTableMapping: filter

    It is often desirable to manage multiple mappings for the same type of
    object and thus same table. The PJ mapping thus supports filtering
    for all its functions.

      >>> class SiteContainer(mapping.PJTableMapping):
      ...     __pj_table__ = 'pjpersist_dot_tests_dot_test_mapping_dot_Item'
      ...     __pj_mapping_key__ = 'name'
      ...     def __init__(self, jar, site):
      ...         super(SiteContainer, self).__init__(jar)
      ...         self.site = site
      ...     def __pj_filter__(self):
      ...         return '''data @> '{"site": "%s"}' ''' % self.site

      >>> container1 = SiteContainer(dm, 'site1')
      >>> container2 = SiteContainer(dm, 'site2')

    Let's now add some items:

      >>> ref11 = dm.dump(Item('1-1', 'site1'))
      >>> ref12 = dm.dump(Item('1-2', 'site1'))
      >>> ref13 = dm.dump(Item('1-3', 'site1'))
      >>> ref21 = dm.dump(Item('2-1', 'site2'))

    And accessing the items works as expected:

      >>> dm.commit(None)

      >>> container1.keys()
      ['1-1', '1-2', '1-3']
      >>> container1['1-1'].name
      '1-1'
      >>> container1['2-1']
      Traceback (most recent call last):
      ...
      KeyError: '2-1'

      >>> container2.keys()
      ['2-1']

    Note: The mutator methods (``__setitem__`` and ``__delitem__``) do nto
    take the filter into account by default. They need to be extended to
    properly setup and tear down the filter criteria.
    """

def test_suite():
    suite = doctest.DocTestSuite(
        setUp=testing.setUp, tearDown=testing.tearDown,
        checker=testing.checker,
        optionflags=testing.OPTIONFLAGS)
    suite.layer = testing.db_layer
    return suite



##############################################################################
#
# Copyright (c) 2014 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Mongo-like queries for PJ"""

import json
from pjpersist import sqlbuilder as sb
from pjpersist import serialize


class Converter(object):
    """Translator for MongoDB queries -> sqlbuilder expressions.

    The resultant expressions can be used as WHERE clauses in queries.

    This implements only a subset of the query language.
    """

    simplified = False

    def __init__(self, table, field):
        self.table = table
        self.field = field

    def convert(self, query):
        clauses = []
        doc = sb.Field(self.table, self.field)
        for key, value in sorted(query.items()):
            accessor = self.getField(doc, key, json=True)
            # some values, esp. datetime must go through PJ serialize
            pjvalue = serialize.ObjectWriter(None).get_state(value)
            jvalue = json.dumps(pjvalue, sort_keys=True)

            if key == '_id':
                jvalue = value

            if key in ('$and', '$or', '$nor', '$startswith'):
                if not isinstance(value, (list, tuple)):
                    raise ValueError("Argument must be a list: %r" % value)
                oper = {
                    '$and': sb.AND,
                    '$or': sb.OR,
                    '$nor': lambda *items: sb.NOT(sb.OR(*items)),
                    '$startswith': sb.STARTSWITH  # special case for a $regex
                }[key]
                clauses.append(oper(*(self.convert(expr) for expr in value)))
            elif isinstance(value, dict):
                for operator, operand in sorted(value.items()):
                    clauses.append(
                        self.operator_expr(
                            operator, doc, key, operand))
            else:
                # Scalar -- equality or array membership
                if self.simplified or key == '_id':
                    # Let's ignore the membership case for test clarity
                    clauses.append(accessor == jvalue)
                else:
                    options = [
                        accessor == jvalue,
                        sb.JSONB_SUBSET(sb.JSONB(json.dumps([pjvalue])),
                                        accessor)
                    ]
                    if value is None:
                        options.append(sb.ISNULL(accessor))
                    clauses.append(sb.OR(*options))
        return sb.AND(*clauses)

    def getField(self, field, key, json=False):
        if isinstance(field, sb.SQLConstant):
            # hack for $elemMatch
            accessor = field
        elif key == '_id':
            accessor = sb.Field(self.table, 'id')
        elif json:
            if '.' not in key:
                accessor = sb.JSON_GETITEM(field, key)
            else:
                accessor = sb.JSON_PATH(field, key.split("."))
        else:
            if '.' not in key:
                accessor = sb.JSON_GETITEM_TEXT(field, key)
            else:
                accessor = sb.JSON_PATH_TEXT(field, key.split("."))

        return accessor

    def operator_expr(self, operator, field, key, op2):
        op1 = self.getField(field, key, json=True)
        # some values, esp. datetime must go through PJ serialize
        pjvalue = serialize.ObjectWriter(None).get_state(op2)
        op2j = json.dumps(pjvalue, sort_keys=True)

        if key == '_id':
            op2j = op2

        if operator == '$gt':
            return op1 > op2j
        if operator == '$lt':
            return op1 < op2j
        if operator == '$gte':
            return op1 >= op2j
        if operator == '$lte':
            return op1 <= op2j
        if operator == '$ne':
            return op1 != op2j
        if operator == '$in':
            if not op2:
                # SQL burps on empty lists with IN:
                return False
            # The $in operator in Mongo is pretty powerful, since it
            # implements the classic "in" (contains) and the "any" operator.
            # There is an optimization for strings:
            #      sb.JSONB_CONTAINS_ANY(op1, [unicode(e) for e in op2])
            ops = [sb.IN(op1, [sb.JSONB(json.dumps(el)) for el in op2])]
            ops += [sb.JSONB_SUPERSET(op1, json.dumps(el)) for el in op2]
            return sb.OR(*ops)
        if operator == '$nin':
            if not op2:
                # SQL burps on empty lists with IN:
                return True
            # The $nin operator in Mongo is pretty powerful, since it
            # implements the classic "not in" (not contains) and the "not any"
            # operator.
            # There is an optimization for strings:
            #      sb.JSONB_CONTAINS_ANY(op1, [unicode(e) for e in op2])
            ops = [sb.IN(op1, [sb.JSONB(json.dumps(el)) for el in op2]),]
            ops += [sb.JSONB_SUPERSET(op1, json.dumps(el)) for el in op2]
            return sb.NOT(sb.OR(*ops))
        if operator == '$not':
            # MongoDB's rationalization for this operator:
            # it matches when op1 does not pass the condition
            # or when op1 is not set at all.
            return sb.OR(
                sb.ISNULL(op1),
                sb.NOT(sb.AND(
                    *(self.operator_expr(operator2, field, key, op3)
                      for operator2, op3 in op2.items())
                )))
        if operator == '$size':
            return sb.func.jsonb_array_length(op1) == op2
        if operator == '$exists':
            return sb.ISNOTNULL(op1) if op2 else sb.ISNULL(op1)
        if operator == '$any':
            return sb.JSONB_CONTAINS_ANY(op1, op2)
        if operator == '$nany':
            return sb.NOT(sb.JSONB_CONTAINS_ANY(op1, op2))
        if operator == '$all':
            return sb.JSONB_SUPERSET(op1, op2j)
        if operator == '$elemMatch':
            op1 = sb.NoTables(op1)
            # SELECT data FROM tbl WHERE EXISTS (
            #          SELECT value
            #          FROM jsonb_array_elements(data -> 'arr')
            #          WHERE value < '3' AND value >= '2'
            # );
            return sb.EXISTS(
                sb.Select(
                    ['value'],
                    staticTables=[sb.func.jsonb_array_elements(op1)],
                    where=sb.NoTables(sb.AND(*(
                        sb.AND(*(
                            self.operator_expr(operator2, sb.SQLConstant('value'), key, op3)
                            for operator2, op3 in query.items()
                        ))
                        for query in op2
                    )))
                )
            )
        if operator == '$startswith':
            return sb.STARTSWITH(sb.TEXT(op1), op2)
        if operator == '$regex':
            return sb.RLIKE(sb.TEXT(op1), op2)
        else:
            raise ValueError("Unrecognized operator %s" % operator)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PJ Data Manager Tests"""
import contextlib
import doctest
import persistent
import unittest
import psycopg2
import psycopg2.errors
from pprint import pprint

import transaction
import mock

from pjpersist import interfaces, serialize, testing, datamanager


class Root(persistent.Persistent):
    pass


class Foo(persistent.Persistent):
    name = None

    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)


class Super(persistent.Persistent):
    _p_pj_table = 'Super'

    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)


class Sub(Super):
    pass


class Bar(persistent.Persistent):
    _p_pj_sub_object = True

    def __init__(self, name=None):
        super(Bar, self).__init__()
        self.name = name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)


class FooItem(object):
    def __init__(self):
        self.bar = 6


class ComplexFoo(persistent.Persistent):
    def __init__(self):
        self.item = FooItem()
        self.name = 'complex'


def doctest_Root():
    r"""Root: General Test

    This class represents the root(s) of the object tree. All roots are stored
    in a specified table. Since the rooted object needs to immediately
    provide a data manager (jar), the operations on the DB root are not art of
    the transaction mechanism.

      >>> root = datamanager.Root(dm, 'proot')

    Initially the root is empty:

      >>> root.keys()
      []

    Let's now add an item:

      >>> foo = Foo()
      >>> root['foo'] = foo
      >>> root.keys()
      ['foo']
      >>> root['foo'] == foo
      True

    Root objects can be overridden:

      >>> foo2 = Foo()
      >>> root['foo'] = foo2
      >>> root.keys()
      ['foo']
      >>> root['foo'] == foo
      False

    And of course we can delete an item:

      >>> del root['foo']
      >>> root.keys()
      []
    """

def doctest_PJDataManager_get_table_from_object():
    r"""PJDataManager: _get_table_from_object(obj)

    Get the table for an object.

      >>> foo = Foo('1')
      >>> foo_ref = dm.insert(foo)

      >>> dbname, table = dm._get_table_from_object(foo)
      >>> dm.reset()

    We are returning the database and table name pair.

      >>> dbname, table
      ('pjpersist_test', 'pjpersist_dot_tests_dot_test_datamanager_dot_Foo')
    """

def doctest_PJDataManager_object_dump_load_reset():
    r"""PJDataManager: dump(), load(), reset()

    The PJ Data Manager is a persistent data manager that manages object
    states in a PostGreSQL database accross Python transactions.

    There are several arguments to create the data manager, but only the
    psycopg2 connection is required:

      >>> dm = datamanager.PJDataManager(
      ...     testing.DummyConnectionPool(conn),
      ...     root_table = 'proot')

    There are two convenience methods that let you serialize and de-serialize
    objects explicitly:

      >>> foo = Foo()
      >>> dm.dump(foo)
      DBRef('pjpersist_dot_tests_dot_test_datamanager_dot_Foo',
            '0001020304050607080a0b0c',
            'pjpersist_test')

    When the object is modified, ``dump()`` will remove it from the list of
    registered objects.

      >>> foo.name = 'Foo'
      >>> foo._p_changed
      True
      >>> list(dm._registered_objects.values())
      [<Foo Foo>]

      >>> foo_ref = dm.dump(foo)

      >>> foo._p_changed
      False
      >>> dm._registered_objects
      {}

      >>> dm.commit(None)

    Let's now reset the data manager, so we do not hit a cache while loading
    the object again:

      >>> dm.reset()

    We can now load the object:

      >>> foo2 = dm.load(foo._p_oid)
      >>> foo == foo2
      False
      >>> foo._p_oid = foo2._p_oid
    """


def doctest_PJDataManager_insertWithExplicitId():
    """
    Objects can be inserted by specifying new object id explicitly.

      >>> foo = Foo('foo')
      >>> foo_ref = dm.insert(foo, '000000000000000000000001')
      >>> dm.tpc_finish(None)

    Now, Foo object should be have the provided id

      >>> foo._p_oid.id
      '000000000000000000000001'
  """


def doctest_PJDataManager_flush():
    r"""PJDataManager: flush()

    This method writes all registered objects to PsotGreSQL. It can be used at
    any time during the transaction when a dump is necessary, but is also used
    at the end of the transaction to dump all remaining objects.

    Let's now add an object to the database and reset the manager like it is
    done at the end of a transaction:

      >>> foo = Foo('foo')
      >>> foo_ref = dm.dump(foo)
      >>> dm.commit(None)

    Let's now load the object again and make a modification:

      >>> foo_new = dm.load(foo._p_oid)
      >>> foo_new.name = 'Foo'

    The object is now registered with the data manager:

      >>> list(dm._registered_objects.values())
      [<Foo Foo>]

    Let's now flush the registered objects:

      >>> dm.flush()

    There are several side effects that should be observed:

    * During a given transaction, we guarantee that the user will always receive
      the same Python object. This requires that flush does not reset the object
      cache.

        >>> id(dm.load(foo._p_oid)) == id(foo_new)
        True

    * The object is removed from the registered objects and the ``_p_changed``
      flag is set to ``False``.

        >>> dm._registered_objects
        {}
        >>> foo_new._p_changed
        False
    """

def doctest_PJDataManager_insert():
    r"""PJDataManager: insert(obj)

    This method inserts an object into the database.

      >>> foo = Foo('foo')
      >>> foo_ref = dm.insert(foo)

    After insertion, the original is not changed:

      >>> foo._p_changed
      False

    It is also added to the list of inserted objects:

      >>> list(dm._inserted_objects.values())
      [<Foo foo>]

    Let's make sure it is really in PostGreSQL:

      >>> dm.commit(None)

      >>> foo_new = dm.load(foo_ref)
      >>> foo_new
      <Foo foo>

    Notice, that we cannot insert the object again:

      >>> dm.insert(foo_new)
      Traceback (most recent call last):
      ...
      ValueError: ('Object._p_oid is already set.', <Foo foo>)

    Finally, registering a new object will not trigger an insert, but only
    schedule the object for writing. This is done, since sometimes objects are
    registered when we only want to store a stub since we otherwise end up in
    endless recursion loops.

      >>> foo2 = Foo('Foo 2')
      >>> dm.register(foo2)

      >>> list(dm._registered_objects.values())
      [<Foo Foo 2>]

    But storing works as expected (flush is implicit before find):

      >>> dm.flush()
      >>> dumpTable(dm._get_table_from_object(foo2)[1])
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'foo'},
        'id': '0001020304050607080a0b0c0'},
       {'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'Foo 2'},
        'id': '0001020304050607080a0b0c0'}]
    """


def doctest_PJDataManager_remove():
    r"""PJDataManager: remove(obj)

    This method removes an object from the database.

      >>> foo = Foo('foo')
      >>> foo_ref = dm.insert(foo)

      >>> dm.commit(None)

    Let's now load the object and remove it.

      >>> foo_new = dm.load(foo_ref)
      >>> dm.remove(foo_new)

    The object is removed from the table immediately:

      >>> dumpTable(dm._get_table_from_object(foo)[1])
      []

    Also, the object is added to the list of removed objects:

      >>> list(dm._removed_objects.values())
      [<Foo foo>]

    Note that you cannot remove objects that are not in the database:

      >>> dm.remove(Foo('Foo 2'))
      Traceback (most recent call last):
      ValueError: ('Object._p_oid is None.', <Foo Foo 2>)

    There is an edge case, if the object is inserted and removed in the same
    transaction:

      >>> dm.commit(None)

      >>> foo3 = Foo('Foo 3')
      >>> foo3_ref = dm.insert(foo3)
      >>> dm.remove(foo3)

    In this case, the object is removed from PostGreSQL and from the inserted
    object list, but it is still added to removed object list, just in case we
    know if it was removed.

      >>> dm._inserted_objects
      {}
      >>> list(dm._removed_objects.values())
      [<Foo Foo 3>]

    """


def doctest_PJDataManager_insert_remove():
    r"""PJDataManager: insert and remove in the same transaction

    Let's insert an object:

      >>> foo = Foo('foo')
      >>> foo_ref = dm.insert(foo)

    And remove it ASAP:

      >>> dm.remove(foo)

      >>> dm._inserted_objects
      {}
      >>> list(dm._removed_objects.values())
      [<Foo foo>]

      >>> dumpTable(dm._get_table_from_object(foo)[1])
      []

    """


def doctest_PJDataManager_insert_remove_modify():
    r"""PJDataManager: insert and remove in the same transaction

    Let's insert an object:

      >>> foo = Foo('foo')
      >>> foo_ref = dm.insert(foo)

    And remove it ASAP:

      >>> dm.remove(foo)

      >>> dm._inserted_objects
      {}
      >>> list(dm._removed_objects.values())
      [<Foo foo>]

      >>> foo.name = 'bar'
      >>> list(dm._removed_objects.values())
      [<Foo bar>]
      >>> list(dm._registered_objects.values())
      []

      >>> dumpTable(dm._get_table_from_object(foo)[1])
      []

      >>> dm.reset()

    """

def doctest_PJDataManager_remove_modify_flush():
    r"""PJDataManager: An object is modified after removal.

    Let's insert an object:

      >>> foo = Foo('foo')
      >>> foo_ref = dm.insert(foo)
      >>> dm.reset()

    Let's now remove it:

      >>> dm.remove(foo)
      >>> list(dm._removed_objects.values())
      [<Foo foo>]

    Within the same transaction we modify the object. But the object should
    not appear in the registered objects list.

      >>> foo._p_changed = True
      >>> dm._registered_objects
      {}

    Now, because of other lookups, the changes are flushed, which should not
    restore the object.

      >>> dm.flush()
      >>> dumpTable(dm._get_table_from_object(foo)[1])
      []
      >>> dm.reset()

    """

def doctest_PJDataManager_remove_flush_modify():
    r"""PJDataManager: An object is removed, DM flushed, object modified

    Let's insert an object:

      >>> foo = Foo('foo')
      >>> foo_ref = dm.insert(foo)
      >>> dm.reset()

    Let's now remove it:

      >>> foo._p_changed = True
      >>> dm.remove(foo)
      >>> list(dm._removed_objects.values())
      [<Foo foo>]

    Now, because of other lookups, the changes are flushed, which should not
    restore the object.

      >>> dm.flush()
      >>> dumpTable(dm._get_table_from_object(foo)[1])
      []

    Within the same transaction we modify the object. But the object should
    not appear in the registered objects list.

      >>> foo._p_changed = True
      >>> dm._registered_objects
      {}

      >>> dumpTable(dm._get_table_from_object(foo)[1])
      []

      >>> dm.reset()

    """


def doctest_PJDataManager_setstate():
    r"""PJDataManager: setstate()

    This method loads and sets the state of an object and joins the
    transaction.

      >>> foo = Foo('foo')
      >>> ref = dm.dump(foo)

      >>> dm.commit(None)
      >>> dm._needs_to_join
      True

      >>> foo2 = Foo()
      >>> foo2._p_oid = ref
      >>> dm.setstate(foo2)
      >>> foo2.name
      'foo'

      >>> dm._needs_to_join
      False
    """


def doctest_PJDataManager_setstate_twice():
    r"""PJDataManager: setstate()

    `setstate` and in turn `set_ghost_state` must not muck with the state
    stored in `_latest_states` otherwise the next setstate will fail badly
    IOW `get_non_persistent_object` must not change it's parameter `state`
    this is a more high level test for the same

      >>> foo = Foo('foo')

      >>> import zope.interface
      >>> ifaces = (zope.interface.Interface, )
      >>> zope.interface.directlyProvides(foo, tuple(ifaces))

      >>> zope.interface.Interface.providedBy(foo)
      True

      >>> ref = dm.dump(foo)

      >>> dm.commit(None)
      >>> dm._needs_to_join
      True

      >>> foo2 = Foo()
      >>> foo2._p_oid = ref
      >>> dm.setstate(foo2)
      >>> foo2.name
      'foo'

      >>> zope.interface.Interface.providedBy(foo2)
      True

      >>> foo3 = Foo()
      >>> foo3._p_oid = ref
      >>> dm.setstate(foo3)
      >>> foo3.name
      'foo'

      >>> zope.interface.Interface.providedBy(foo3)
      True
    """


def doctest_PJDataManager_oldstate():
    r"""PJDataManager: oldstate()

    Loads the state of an object for a given transaction. Since we are not
    supporting history, this always raises a key error as documented.

      >>> foo = Foo('foo')
      >>> dm.oldstate(foo, '0')
      Traceback (most recent call last):
      ...
      KeyError: '0'
    """

def doctest_PJDataManager_register():
    r"""PJDataManager: register()

    Registers an object to be stored.

      >>> dm.reset()
      >>> dm._needs_to_join
      True
      >>> len(dm._registered_objects)
      0

      >>> foo = Foo('foo')
      >>> dm.register(foo)

      >>> dm._needs_to_join
      False
      >>> len(dm._registered_objects)
      1

   But there are no duplicates:

      >>> dm.register(foo)
      >>> len(dm._registered_objects)
      1
    """

def doctest_PJDataManager_abort():
    r"""PJDataManager: abort()

    Aborts a transaction, which clears all object and transaction registrations:

      >>> foo = Foo()
      >>> dm._registered_objects = {id(foo): foo}
      >>> dm._needs_to_join = False

      >>> dm.abort(transaction.get())

      >>> dm._needs_to_join
      True
      >>> len(dm._registered_objects)
      0

    Let's now create a more interesting case with a transaction that inserted,
    removed and changed objects.

    First let's create an initial state:

      >>> dm.reset()
      >>> foo_ref = dm.insert(Foo('one'))
      >>> foo2_ref = dm.insert(Foo('two'))
      >>> dm.commit(None)

      >>> dbanme, table = dm._get_table_from_object(Foo())
      >>> dumpTable(table)
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'one'},
        'id': '0001020304050607080a0b0c0'},
       {'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'two'},
        'id': '0001020304050607080a0b0c0'}]

    Now, in a second transaction we modify the state of objects in all three
    ways:

      >>> foo = dm.load(foo_ref)
      >>> foo.name = '1'
      >>> list(dm._registered_objects.values())
      [<Foo 1>]

      >>> foo2 = dm.load(foo2_ref)
      >>> dm.remove(foo2)
      >>> list(dm._removed_objects.values())
      [<Foo two>]

      >>> foo3_ref = dm.insert(Foo('three'))

      >>> dm.flush()
      >>> dumpTable(table)
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': '1'},
        'id': '0001020304050607080a0b0c0'},
       {'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'three'},
        'id': '0001020304050607080a0b0c0'}]

    Let's now abort the transaction and everything should be back to what it
    was before:

      >>> dm.abort(transaction.get())
      >>> dumpTable(table)
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'one'},
        'id': '0001020304050607080a0b0c0'},
       {'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'two'},
        'id': '0001020304050607080a0b0c0'}]
    """


def doctest_PJDataManager_abort_subobjects():
    r"""PJDataManager: abort(): Correct restoring of complex objects

    Object, that contain subobjects should be restored to the state, exactly
    matching one before initial loading.

    1. Create a single record and make sure it is stored in db

      >>> dm.reset()
      >>> foo1_ref = dm.insert(ComplexFoo())
      >>> dm.commit(None)

      >>> dbname, table = dm._get_table_from_object(ComplexFoo())
      >>> dumpTable(table)
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.ComplexFoo',
                 'item': {'_py_type': 'pjpersist.tests.test_datamanager.FooItem',
                           'bar': 6},
                 'name': 'complex'},
        'id': '0001020304050607080a0b0c0'}]

    2. Modify the item and flush it to database

      >>> foo1 = dm.load(foo1_ref)
      >>> foo1.name = 'modified'
      >>> dm.flush()

      >>> dumpTable(table)
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.ComplexFoo',
                 'item': {'_py_type': 'pjpersist.tests.test_datamanager.FooItem',
                           'bar': 6},
                 'name': 'modified'},
        'id': '0001020304050607080a0b0c0'}]

    3. Abort the current transaction and expect original state is restored

      >>> dm.abort(transaction.get())
      >>> dumpTable(table)
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.ComplexFoo',
                 'item': {'_py_type': 'pjpersist.tests.test_datamanager.FooItem',
                           'bar': 6},
                 'name': 'complex'},
        'id': '0001020304050607080a0b0c0'}]
    """

def doctest_PJDataManager_tpc_begin():
    r"""PJDataManager: tpc_begin()

    This is a non-op for the PJ data manager.

      >>> dm.tpc_begin(transaction.get())
    """

def doctest_PJDataManager_tpc_vote():
    r"""PJDataManager: tpc_vote()

    This is a non-op for the PJ data manager.

      >>> dm.tpc_vote(transaction.get())
    """

def doctest_PJDataManager_tpc_finish():
    r"""PJDataManager: tpc_finish()

    This method finishes the two-phase commit. Let's store a simple object:

      >>> foo = Foo()
      >>> dm.register(foo)
      >>> transaction.commit()

    Note that objects cannot be stored twice in the same transaction:

      >>> dm.reset()
      >>> dm.register(foo)
      >>> dm.register(foo)
      >>> transaction.commit()

    Also, when a persistent sub-object is stored that does not want its own
    document, then its parent is stored instead, still avoiding dual storage.

      >>> dm.reset()
      >>> foo2 = dm.load(foo._p_oid)
      >>> foo2.bar = Bar()
    """

def doctest_PJDataManager_tpc_abort():
    r"""PJDataManager: tpc_abort()

    Aborts a two-phase commit. This is simply the same as the regular abort.

      >>> foo = Foo()
      >>> dm._registered_objects = {id(foo): foo}
      >>> dm._needs_to_join = False

      >>> dm.tpc_abort(transaction.get())

      >>> dm._needs_to_join
      True
      >>> len(dm._registered_objects)
      0
    """


def doctest_PJDataManager_transaction_abort_after_query():
    r"""

    When we perform illegal sql, connection is set to "aborted" state, and you
    cannot execute any more queries on it. However, after you abort the
    transaction, you can continue.

    Let's execute bad SQL

      >>> foo = Foo()
      >>> cur = dm.getCursor()
      >>> try:
      ...     cur.execute("SELECT 1/0")
      ...     cur.fetchall()
      ... except:
      ...     transaction.abort()

    We aborted transaction and now we can continue doing stuff

      >>> cur = dm.getCursor()
      >>> cur.execute("SELECT 1")
      >>> cur.fetchall()
      [[1]]

    """


def doctest_PJDataManager_sortKey():
    r"""PJDataManager: sortKey()

    The data manager's sort key is trivial.

      >>> dm.sortKey()
      'PJDataManager:...'
    """


def doctest_PJDataManager_sub_objects():
    r"""PJDataManager: Properly handling initialization of sub-objects.

    When `_p_pj_sub_object` objects are loaded from PostGreSQL, their `_p_jar`
    and more importantly their `_p_pj_doc_object` attributes are
    set.

    However, when a sub-object is initially added, those attributes are
    missing.

      >>> foo = Foo('one')
      >>> dm.root['one'] = foo
      >>> commit()

      >>> foo = dm.root['one']
      >>> foo._p_changed

      >>> foo.list = serialize.PersistentList()
      >>> foo.list._p_jar
      >>> getattr(foo.list, '_p_pj_doc_object', 'Missing')
      'Missing'

    Of course, the parent object has changed, since an attribute has been set
    on it.

      >>> foo._p_changed
      True

    Now, since we are dealing with an external database and queries, it
    frequently happens that all changed objects are flushed to the database
    before running a query. In our case, this saves the main object andmarks
    it unchanged again:

      >>> dm.flush()
      >>> foo._p_changed
      False

    However, while flushing, no object is read from the database again.  If
    the jar and document obejct are not set on the sub-object, any changes to
    it would not be seen. Thus, the serialization process *must* assign the
    jar and document object attributes, if not set.

      >>> foo.list._p_jar is dm
      True
      >>> foo.list._p_pj_doc_object is foo
      True

    Let's now ensure that changing the sub-object will have the proper effect:

      >>> foo.list.append(1)
      >>> foo.list._p_changed
      True
      >>> commit()

      >>> foo = dm.root['one']
      >>> foo.list
      [1]

    Note: Most of the implementation of this feature is in the `getState()`
    method of the `ObjectWriter` class.
    """


def doctest_PJDataManager_sub_objects_add_modify():
    """PJDataManager: Make sure that subobject modifications get noticed
    after the first object add

      >>> foo = Foo('one')
      >>> bar = Bar('bar')

      >>> bar._p_pj_sub_object = True
      >>> bar._p_pj_doc_object = foo
      >>> foo.bar = bar

      >>> dm.root['one'] = foo
      >>> commit()

    Now change a subobject property

      >>> bar.name = 'new'
      >>> commit()

    And reload from the DB:

      >>> dm.root['one'].bar.name
      'new'
"""


def doctest_PJDataManager_complex_sub_objects():
    """PJDataManager: Never store objects marked as _p_pj_sub_object

    Let's construct complex object with several levels of containment.
    _p_pj_doc_object will point to an object, that is subobject itself.

      >>> foo = Foo('one')
      >>> sup = Super('super')
      >>> bar = Bar('bar')

      >>> bar._p_pj_sub_object = True
      >>> bar._p_pj_doc_object = sup
      >>> sup.bar = bar

      >>> sup._p_pj_sub_object = True
      >>> sup._p_pj_doc_object = foo
      >>> foo.sup = sup

      >>> dm.root['one'] = foo
      >>> commit()

      >>> cur = conn.cursor()
      >>> cur.execute('SELECT tablename from pg_tables;')
      >>> sorted(e[0] for e in cur.fetchall()
      ...        if not e[0].startswith('pg_') and not e[0].startswith('sql_'))
      ['persistence_root',
       'pjpersist_dot_tests_dot_test_datamanager_dot_foo']

    Now, save foo first, and then add subobjects

      >>> foo = Foo('two')
      >>> dm.root['two'] = foo
      >>> commit()

      >>> sup = Super('second super')
      >>> bar = Bar('second bar')

      >>> bar._p_pj_sub_object = True
      >>> bar._p_pj_doc_object = sup
      >>> sup.bar = bar

      >>> sup._p_pj_sub_object = True
      >>> sup._p_pj_doc_object = foo
      >>> foo.sup = sup
      >>> commit()

      >>> cur.execute('SELECT tablename from pg_tables;')
      >>> sorted(e[0] for e in cur.fetchall()
      ...        if not e[0].startswith('pg_') and not e[0].startswith('sql_'))
      ['persistence_root',
       'pjpersist_dot_tests_dot_test_datamanager_dot_foo']

      >>> dm.root['two'].sup.bar
      <Bar second bar>

      >>> cur = dm.getCursor()
      >>> cur.execute(
      ... '''SELECT * FROM pjpersist_dot_tests_dot_test_datamanager_dot_foo
      ...    WHERE data @> '{"name": "one"}' ''')
      >>> pprint([dict(e) for e in cur.fetchall()])
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Foo',
                 'name': 'one',
                 'sup': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Super',
                          'bar': {'_py_persistent_type': 'pjpersist.tests.test_datamanager.Bar',
                                   'name': 'bar'},
                          'name': 'super'}},
        'id': '0001020304050607080a0b0c0'}]

    Now, make changes to the subobjects and then commit

      >>> foo = dm.root['one']
      >>> foo.sup.name = 'new super'
      >>> foo.sup.bar.name = 'new bar'
      >>> commit()

      >>> foo = dm.root['one']
      >>> foo.sup
      <Super new super>
      >>> foo.sup._p_pj_sub_object
      True
      >>> foo.sup._p_pj_doc_object
      <Foo one>

      >>> foo.sup.bar
      <Bar new bar>

      >>> foo.sup.bar._p_pj_sub_object
      True
      >>> foo.sup.bar._p_pj_doc_object
      <Foo one>

      >>> cur.execute('SELECT tablename from pg_tables;')
      >>> sorted(e[0] for e in cur.fetchall()
      ...        if not e[0].startswith('pg_') and not e[0].startswith('sql_'))
      ['persistence_root',
       'pjpersist_dot_tests_dot_test_datamanager_dot_foo']

    Even if _p_pj_doc_object is pointed to subobject, subobject does not get
    saved to its own table:

      >>> foo.sup.bar._p_pj_doc_object = foo.sup
      >>> foo.sup.bar.name = 'newer bar'
      >>> foo.sup.name = 'newer sup'
      >>> commit()

      >>> cur.execute('SELECT tablename from pg_tables;')
      >>> sorted(e[0] for e in cur.fetchall()
      ...        if not e[0].startswith('pg_') and not e[0].startswith('sql_'))
      ['persistence_root',
       'pjpersist_dot_tests_dot_test_datamanager_dot_foo']
    """


def doctest_PJDataManager_table_sharing():
    r"""PJDataManager: Properly share tables with sub-classes

    When objects do not specify a table, then a table based on the
    class path is created for them. In that case, when a sub-class is created,
    the same table should be used. However, during de-serialization, it
    is important that we select the correct class to use.

      >>> dm.root['app'] = Root()

      >>> dm.root['app'].one = Super('one')
      >>> dm.root['app'].one
      <Super one>

      >>> dm.root['app'].two = Sub('two')
      >>> dm.root['app'].two
      <Sub two>

      >>> dm.root['app'].three = Sub('three')
      >>> dm.root['app'].three
      <Sub three>

      >>> transaction.commit()

    Let's now load everything again:

      >>> dm.root['app'].one
      <Super one>
      >>> dm.root['app'].two
      <Sub two>
      >>> dm.root['app'].three
      <Sub three>
      >>> transaction.commit()

    Make sure that after a restart, the objects can still be stored.

      >>> serialize.AVAILABLE_NAME_MAPPINGS = set()
      >>> serialize.PATH_RESOLVE_CACHE = {}
      >>> dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn))

      >>> dm2.root['app'].four = Sub('four')
      >>> transaction.commit()

      >>> serialize.AVAILABLE_NAME_MAPPINGS = set()
      >>> serialize.PATH_RESOLVE_CACHE = {}

      >>> dm2.root['app'].four
      <Sub four>
    """


def doctest_PJDataManager_no_compare():
    r"""PJDataManager: No object methods are called during register/dump.

    Using object comparison within the data manager can have undesired side
    effects. For example, `__cmp__()` could make use of other model objects
    that cause flushes and queries in the data manager. This can have very
    convoluted side effects, including loss of data.

      >>> class BadObject(persistent.Persistent):
      ...     def __init__(self, name):
      ...         self.name = name
      ...     def __cmp__(self, other):
      ...         raise ValueError('Compare used in data manager!!!')
      ...     def __repr__(self):
      ...         return '<BadObject %s>' % self.name

      >>> dm.root['bo1'] = BadObject('bo1')
      >>> dm.root['bo2'] = BadObject('bo2')

      >>> dm.tpc_finish(None)

    Since `__cmp__()` was not used, no exception was raised.

      >>> bo1 = dm.root['bo1']
      >>> bo1
      <BadObject bo1>
      >>> bo2 = dm.root['bo2']
      >>> bo2
      <BadObject bo2>

      >>> dm.register(bo1)
      >>> dm.register(bo2)
      >>> sorted(dm._registered_objects.values(), key=lambda ob: ob.name)
      [<BadObject bo1>, <BadObject bo2>]

    """

def doctest_PJDataManager_modify_sub_delete_doc():
    """PJDataManager: Deletion is not cancelled if sub-object is modified.

    It must be ensured that the deletion of an object is not cancelled when a
    sub-document object is modified (since it is registered with the data
    manager.

      >>> foo = Foo('foo')
      >>> dm.root['foo'] = foo
      >>> foo.bar = Bar('bar')

      >>> dm.tpc_finish(None)
      >>> cur = dm.getCursor()
      >>> cur.execute(
      ...     '''SELECT count(*)
      ...        FROM pjpersist_dot_tests_dot_test_datamanager_dot_Foo''')
      >>> cur.fetchone()[0]
      1L

    Let's now modify bar and delete foo.

      >>> foo = dm.root['foo']
      >>> foo.bar.name = 'bar-new'
      >>> dm.remove(foo)

      >>> dm.tpc_finish(None)
      >>> cur.execute(
      ...     '''SELECT count(*)
      ...        FROM pjpersist_dot_tests_dot_test_datamanager_dot_Foo''')
      >>> cur.fetchone()[0]
      0L
    """

def doctest_PJDataManager_sub_doc_multi_flush():
    """PJDataManager: Sub-document object multi-flush

    Make sure that multiple changes to the sub-object are registered, even if
    they are flushed inbetween. (Note that flushing happens often due to
    querying.)

      >>> foo = Foo('foo')
      >>> dm.root['foo'] = foo
      >>> foo.bar = Bar('bar')

      >>> commit()

    Let's now modify bar a few times with intermittend flushes.

      >>> foo = dm.root['foo']
      >>> foo.bar.name = 'bar-new'
      >>> dm.flush()
      >>> foo.bar.name = 'bar-newer'

      >>> commit()
      >>> dm.root['foo'].bar.name
      'bar-newer'
    """


def doctest_get_database_name_from_dsn():
    """Test dsn parsing

      >>> from pjpersist.datamanager import get_database_name_from_dsn

      >>> get_database_name_from_dsn("dbname=test user=postgres password=secret")
      'test'

      >>> get_database_name_from_dsn("dbname = test  user='postgres'")
      'test'

      >>> get_database_name_from_dsn("user='postgres' dbname = test")
      'test'

      >>> get_database_name_from_dsn("user='pg' dbname =test   password=pass")
      'test'
    """


class DatamanagerConflictTest(testing.PJTestCase):

    def test_conflict_del_1(self):
        """Check conflict detection. We modify and delete the same object in
        different transactions, simulating separate processes."""

        txn = transaction.manager.get()
        foo = Foo('foo-first')
        self.dm.root['foo'] = foo

        transaction.commit()

        conn1 = testing.getConnection(testing.DBNAME)
        conn2 = testing.getConnection(testing.DBNAME)

        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))
        self.assertEqual(dm2.root['foo'].name, 'foo-first')
        del dm2.root['foo']

        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))
        self.assertEqual(dm1.root['foo'].name, 'foo-first')
        dm1.root['foo'].name = 'foo-second'

        #Finish in order 2 - 1

        with self.assertRaises(interfaces.ConflictError):
            transaction.commit()

        transaction.abort()

        conn2.close()
        conn1.close()

    def test_conflict_del_2(self):
        """Check conflict detection. We modify and delete the same object in
        different transactions, simulating separate processes."""

        foo = Foo('foo-first')
        self.dm.root['foo'] = foo

        transaction.commit()

        conn1 = testing.getConnection(testing.DBNAME)
        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))

        self.assertEqual(dm1.root['foo'].name, 'foo-first')

        dm1.root['foo'].name = 'foo-second'

        conn2 = testing.getConnection(testing.DBNAME)
        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))

        self.assertEqual(dm2.root['foo'].name, 'foo-first')
        del dm2.root['foo']

        #Finish in order 1 - 2
        # well, try to... dm1.tpc_finish will block until dm2 is done

        @testing.run_in_thread
        def background_commit():
            with self.assertRaises(interfaces.ConflictError):
                dm1.commit(None)
        dm2.commit(None)

        transaction.abort()

        conn2.close()
        conn1.close()

    def test_conflict_tracebacks(self):
        """Verify conflict tracebacks are captured properly
        and reset on the next transaction."""

        ctb = datamanager.CONFLICT_TRACEBACK_INFO.traceback
        self.assertIsNone(ctb)

        foo = Foo('foo-first')
        self.dm.root['foo'] = foo

        transaction.commit()

        conn2 = testing.getConnection(testing.DBNAME)
        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))
        del dm2.root['foo']

        conn1 = testing.getConnection(testing.DBNAME)
        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))
        dm1.root['foo'].name = 'foo-second'

        ctb = datamanager.CONFLICT_TRACEBACK_INFO.traceback
        self.assertIsNone(ctb)

        # Finish in order 2 - 1
        with self.assertRaises(interfaces.ConflictError):
            transaction.commit()

        # verify by length that we have the full traceback
        ctb = datamanager.CONFLICT_TRACEBACK_INFO.traceback
        self.assertIsNotNone(ctb)
        # Make all work
        self.assertTrue(len(ctb) > 20)
        self.assertIn('Beacon:', ctb[-1])
        transaction.abort()

        # start another transaction and verify the traceback
        # is reset
        datamanager.PJDataManager(testing.DummyConnectionPool(conn2))

        ctb = datamanager.CONFLICT_TRACEBACK_INFO.traceback
        self.assertIsNone(ctb)

        conn2.close()
        conn1.close()

    def test_conflict_commit_1(self):
        """Test conflict on commit

        The typical detail string for such failures is:

        DETAIL:  Reason code: Canceled on identification as a pivot, during
        commit attempt.
        """

        # We will not reproduce the full scenario with pjpersist, however we
        # will pretend the right exception is thrown by commit.
        #
        # First, get the error, that psycopg throws in such case
        # The example is taken from https://wiki.postgresql.org/wiki/SSI
        conn1 = self.conn

        conn2 = testing.getConnection(testing.DBNAME)

        with conn1.cursor() as cur:
            cur.execute("DROP TABLE IF EXISTS mytab")
            cur.execute(
                "CREATE TABLE mytab (class int NOT NULL, value int NOT NULL )")
            cur.execute(
                "INSERT INTO mytab VALUES (1, 10), (1, 20), (2, 100), (2, 200)")
        conn1.commit()

        with self.dm.getCursor() as cur1, conn2.cursor() as cur2:
            cur1.execute("SELECT SUM(value) FROM mytab WHERE class = 1")
            cur1.execute("INSERT INTO mytab VALUES (2, 30)")

            cur2.execute("SELECT SUM(value) FROM mytab WHERE class = 2")
            cur2.execute("INSERT INTO mytab VALUES (1, 300)")

        conn2.commit()
        conn2.close()

        # Now datamanager, holding conn1 is in doomed state. it is expected to
        # fail on commit attempt.
        with self.assertRaises(interfaces.ConflictError):
            transaction.commit()

    def test_catch_bad_sql(self):
        """Test breaks psycopg2 transaction but tries to commit anyway."""

        # An offending piece of code swallows psycopg2 exception,
        # leaving psycopg2 transaction in a failed state
        with self.dm.getCursor() as cursor:
            with self.assertRaises(psycopg2.errors.SyntaxError):
                cursor.execute("BAD SQL")

        # The transaction is now marked as doomed, so we cannot commit it
        with self.assertRaises(transaction.interfaces.DoomedTransaction):
            transaction.commit()
        transaction.abort()

    def test_db_disconnect(self):
        """check_for_disconnect converts some psycopg2 exceptions to
        DatabaseDisconnected exceptions, check that
        """
        def fail(*args, **kw):
            raise psycopg2.OperationalError('boom')

        with mock.patch(
                'pjpersist.datamanager.PJPersistCursor._execute_and_log',
                side_effect=fail):
            with self.dm.getCursor() as cur:
                with self.assertRaises(interfaces.DatabaseDisconnected):
                    cur.execute("DROP TABLE IF EXISTS mytab")

        def fail(*args, **kw):
            raise psycopg2.InterfaceError('boom')

        with mock.patch(
                'pjpersist.datamanager.PJPersistCursor._execute_and_log',
                side_effect=fail):
            with self.dm.getCursor() as cur:
                with self.assertRaises(interfaces.DatabaseDisconnected):
                    cur.execute("DROP TABLE IF EXISTS mytab")


class QueryLoggingTestCase(testing.PJTestCase):
    def setUp(self):
        super(QueryLoggingTestCase, self).setUp()
        self.log = testing.setUpLogging(datamanager.TABLE_LOG)

        with self.conn.cursor() as cur:
            cur.execute("DROP TABLE IF EXISTS mytab")
            cur.execute("CREATE TABLE mytab (class int NOT NULL, value varchar NOT NULL )")

        pjal_patch = mock.patch("pjpersist.datamanager.PJ_ACCESS_LOGGING",
                                True)
        self.patches = [pjal_patch]
        for p in self.patches:
            p.start()

    def tearDown(self):
        for p in self.patches:
            p.stop()

        super(QueryLoggingTestCase, self).tearDown()
        testing.tearDownLogging(datamanager.TABLE_LOG)

    def test_logging(self):
        with self.dm.getCursor() as cur:
            cur.execute("INSERT INTO mytab VALUES (1, '10')")

        lines = self.log.getvalue().split('\n')
        self.assertEqual(lines[0], "INSERT INTO mytab VALUES (1, '10'),")
        self.assertEqual(lines[1], " args:None,")

    def test_params(self):
        with self.dm.getCursor() as cur:
            cur.execute("INSERT INTO mytab VALUES (%s, %s)", [1, '10'])

        lines = self.log.getvalue().split('\n')
        self.assertEqual(lines[0], "INSERT INTO mytab VALUES (%s, %s),")
        self.assertEqual(lines[1], " args:[1, '10'],")

    def test_long_params(self):
        hugeparam = "1234567890" * 20000
        with self.dm.getCursor() as cur:
            cur.execute("INSERT INTO mytab VALUES (%s, %s)", [1, hugeparam])

        lines = self.log.getvalue().split('\n')
        self.assertEqual(lines[0], "INSERT INTO mytab VALUES (%s, %s),")
        self.assertLess(len(lines[1]), 1000)


class TransactionOptionsTestCase(testing.PJTestCase):
    def setUp(self):
        super(TransactionOptionsTestCase, self).setUp()

        # Transaction options feature isn't really compatible with table
        # autocreation, because transaction features has to be set before any
        # statement is executed in transaction. So we turn it off in these
        # tests.
        pjact_patch = mock.patch("pjpersist.datamanager.PJ_AUTO_CREATE_TABLES",
                                 False)
        pjacc_patch = mock.patch("pjpersist.datamanager.PJ_AUTO_CREATE_COLUMNS",
                                 False)
        self.patches = [pjact_patch, pjacc_patch]
        for p in self.patches:
            p.start()

        with self.conn.cursor() as cur:
            cur.execute("DROP TABLE IF EXISTS mytab")
            cur.execute("CREATE TABLE mytab (class int NOT NULL, value varchar NOT NULL )")
        transaction.commit()
        self.dm.reset()

    def tearDown(self):
        for p in self.patches:
            p.stop()

        super(TransactionOptionsTestCase, self).tearDown()

    def test_setTransactionOptions_setIsolation(self):
        """It is possible to request transaction options before first
        statement is executed
        """

        self.dm.setTransactionOptions(isolation_level="READ COMMITTED")

        cur = self.dm.getCursor()
        cur.execute('SHOW transaction_isolation')
        res = cur.fetchone()
        self.assertEqual(res[0], 'read committed')

    def test_option_reset_after_commit(self):
        # First, check the default isolation level
        cur = self.dm.getCursor()
        cur.execute('SHOW transaction_isolation')
        res = cur.fetchone()
        default_level = res[0]
        self.assertNotEqual(default_level, 'repeatable read')
        transaction.commit()

        # Now change it
        self.dm.setTransactionOptions(isolation_level="REPEATABLE READ")
        cur = self.dm.getCursor()
        cur.execute('SHOW transaction_isolation')
        res = cur.fetchone()
        self.assertEqual(res[0], 'repeatable read')
        transaction.commit()

        # On the subsequent transaction, we should go back to default level
        cur = self.dm.getCursor()
        cur.execute('SHOW transaction_isolation')
        res = cur.fetchone()
        self.assertEqual(res[0], default_level)


class DirtyTestCase(testing.PJTestCase):

    def setUp(self):
        super(DirtyTestCase, self).setUp()

        # get rid of the previous transaction
        transaction.abort()

        tpc_patch = mock.patch(
            "pjpersist.datamanager.PJ_TWO_PHASE_COMMIT_ENABLED", True)
        no_prep_patch = mock.patch(
            "pjpersist.datamanager."
            "CALL_TPC_PREPARE_ON_NO_WRITE_TRANSACTION", False)
        log_patch = mock.patch(
            "pjpersist.datamanager.LOG_READ_WRITE_TRANSACTION", True)
        self.patches = [tpc_patch, no_prep_patch, log_patch]
        for p in self.patches:
            p.start()

        # First PJDataManager instantiation creates tables, what makes the dm
        # dirty, which we want to avoid here.
        self.conn = testing.getConnection(testing.DBNAME)
        self.dm = datamanager.PJDataManager(testing.DummyConnectionPool(self.conn))

    def tearDown(self):
        for p in self.patches:
            p.stop()

        super(DirtyTestCase, self).tearDown()

    def test_isDirty(self):
        """Test PJDataManager.isDirty setting
        """
        # by default a pristine DM is not dirty
        self.assertEqual(self.dm.isDirty(), False)

        with self.cursor():
            # add an object
            self.dm.root['foo'] = Foo('foo-first')
            self.assertEqual(self.dm.isDirty(), True)

        # a commit/abort clears the dirty flag
        self.assertEqual(self.dm.isDirty(), False)

        with self.cursor(abort=True):
            # modify an object property
            self.dm.root['foo'].name = 'blabla'
            self.assertEqual(self.dm.isDirty(), True)

        # a commit/abort clears the dirty flag
        self.assertEqual(self.dm.isDirty(), False)

        # delete an object in a separate transaction
        with self.cursor():
            self.dm.root['foo2'] = Foo('foo-second')

        with self.cursor():
            del self.dm.root['foo2']
            self.assertEqual(self.dm.isDirty(), True)

        # add and remove an object in the same transaction
        with self.cursor():
            self.dm.root['foo3'] = Foo('foo-third')
            del self.dm.root['foo3']
            self.assertEqual(self.dm.isDirty(), True)

        # check the special dump method
        with self.cursor():
            self.dm.dump(self.dm.root['foo'])
            self.assertEqual(self.dm.isDirty(), True)

    def test_isDirty_sql(self):
        """Test PJDataManager.isDirty setting, SQL commands need to set it too
        """
        with self.cursor() as cur:
            cur.execute("SELECT 1")
            self.assertFalse(self.dm.isDirty())

        with self.cursor() as cur:
            cur.execute("INSERT INTO mytab VALUES (%s, %s)", [1, '10'])
            self.assertTrue(self.dm.isDirty())

        # commit clears dirty
        self.assertFalse(self.dm.isDirty())

        with self.cursor() as cur:
            cur.execute("DROP TABLE IF EXISTS mytab")
            cur.execute("CREATE TABLE mytab "
                        "(class int NOT NULL, value varchar NOT NULL )")
            self.assertTrue(self.dm.isDirty())

        with self.cursor() as cur:
            cur.execute("INSERT INTO mytab VALUES (%s, %s)", [1, '10'])

        with self.cursor(abort=True) as cur:
            cur.execute("UPDATE mytab SET value='42' WHERE class=1")
            self.assertTrue(self.dm.isDirty())

        self.assertFalse(self.dm.isDirty())

    def test_clean_dm_tpc_no_prepare(self):
        # check that a DM that has NO writes does NOT call tpc_prepare
        self.assertFalse(self.dm.isDirty())

        # load somehing from the DB to get going with PJDataManager
        self.assertEqual(len(self.dm.root), 0)

        # cannot mock patch a C method (tpc_prepare)
        # next best is _might_execute_with_error
        orig = self.dm._might_execute_with_error
        prep_mock = mock.patch.object(
            self.dm, '_might_execute_with_error', side_effect=orig)
        with prep_mock as prep_mock_p:
            transaction.commit()

        self.assertEqual(prep_mock_p.call_count, 1)  # tpc_commit in tpc_finish
        self.assertFalse('tpc_prepare' in str(prep_mock_p.call_args_list))
        self.assertTrue('tpc_commit' in str(prep_mock_p.call_args_list))

    def test_dirty_dm_tpc_prepare(self):
        # check that a DM that HAS writes DOES call tpc_prepare

        # add an object, do some changes to make the DM dirty
        self.dm.root['foo'] = Foo('foo-first')
        self.assertTrue(self.dm.isDirty())

        # cannot mock patch a C method (tpc_prepare)
        orig = self.dm._might_execute_with_error
        prep_mock = mock.patch.object(
            self.dm, '_might_execute_with_error', side_effect=orig)
        with prep_mock as prep_mock_p:
            transaction.commit()

        self.assertEqual(prep_mock_p.call_count, 2)  # tpc_commit in tpc_finish
        self.assertTrue('tpc_prepare' in str(prep_mock_p.call_args_list))
        self.assertTrue('tpc_commit' in str(prep_mock_p.call_args_list))

    def test_release_on_destroy(self):
        # When DM is destroyed without committing or aborting the transaction
        # (for whatever awkward reason), connection should be released.
        conn = testing.getConnection(testing.DBNAME)
        pool = testing.DummyConnectionPool(conn)
        dm = datamanager.PJDataManager(pool)
        dm.getCursor()  # join the transaction
        self.assertTrue(pool.isTaken())

        # WHEN
        # simulate the garbage collection
        dm.__del__()

        # THEN
        self.assertFalse(pool.isTaken())

    def test_release_on_new_transaction(self):
        # When new transaction is started, connection is released because
        # the previous one is aborted.
        # GIVEN
        conn = testing.getConnection(testing.DBNAME)
        pool = testing.DummyConnectionPool(conn)
        dm = datamanager.PJDataManager(pool)
        dm.getCursor()  # join the transaction
        self.assertTrue(pool.isTaken())

        # WHEN

        # Create new transaction to release the reference to our data manager
        # and allow it to be destroyed
        transaction.begin()

        # THEN
        self.assertFalse(pool.isTaken())

    def test_release_on_commit(self):
        # GIVEN
        transaction.abort()
        pool = testing.DummyConnectionPool(self.conn)
        dm = datamanager.PJDataManager(pool)
        dm.getCursor()  # join the transaction
        self.assertTrue(pool.isTaken())

        # WHEN

        # Create new transaction to release the reference to our data manager
        # and allow it to be destroyed
        transaction.commit()

        # THEN
        self.assertFalse(pool.isTaken())

    def test_release_on_abort(self):
        # GIVEN
        transaction.abort()
        pool = testing.DummyConnectionPool(self.conn)
        dm = datamanager.PJDataManager(pool)
        dm.getCursor()  # join the transaction
        self.assertTrue(pool.isTaken())

        # WHEN

        # Create new transaction to release the reference to our data manager
        # and allow it to be destroyed
        transaction.abort()

        # THEN
        self.assertFalse(pool.isTaken())

    def test_readonly_fail_on_write(self):
        transaction.abort()
        pool = testing.DummyConnectionPool(self.conn)
        dm = datamanager.PJDataManager(pool)
        dm.setTransactionOptions(readonly=True)

        # WHEN, THEN
        with self.assertRaises(interfaces.ReadOnlyDataManagerError):
            with dm.getCursor() as cur:
                cur.execute("INSERT INTO mytab VALUES (1, '10')")

    @contextlib.contextmanager
    def cursor(self, abort=False):
        transaction.abort()
        self.dm = datamanager.PJDataManager(testing.DummyConnectionPool(self.conn))
        cur = self.dm.getCursor()
        yield cur
        if abort:
            transaction.abort()
        else:
            transaction.commit()


def test_suite():
    dtsuite = doctest.DocTestSuite(
        setUp=testing.setUp, tearDown=testing.tearDown,
        checker=testing.checker,
        optionflags=testing.OPTIONFLAGS)
    dtsuite.layer = testing.db_layer

    return unittest.TestSuite((
        dtsuite,
        unittest.makeSuite(DatamanagerConflictTest),
        unittest.makeSuite(QueryLoggingTestCase),
        unittest.makeSuite(TransactionOptionsTestCase),
        unittest.makeSuite(DirtyTestCase),
        ))



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PG/JSONB Persistence Interfaces"""
import datetime
import decimal
import persistent.interfaces
import transaction.interfaces
import types
import zope.interface
import zope.schema

PJ_NATIVE_TYPES = (int, bool, float, str, type(None))

REFERENCE_SAFE_TYPES = (
    datetime.datetime, datetime.date, datetime.time, decimal.Decimal)

DATABASE_ATTR_NAME = '_p_pj_database'
TABLE_ATTR_NAME = '_p_pj_table'
SUB_OBJECT_ATTR_NAME = '_p_pj_sub_object'
DOC_OBJECT_ATTR_NAME = '_p_pj_doc_object'
STATE_ATTR_NAME = '_p_pj_state'
PY_TYPE_ATTR_NAME = '_py_persistent_type'


class CircularReferenceError(Exception):
    pass


class ConflictError(transaction.interfaces.TransientError):
    pass


class DatabaseDisconnected(transaction.interfaces.TransientError):
    pass


class ReadOnlyDataManagerError(transaction.interfaces.TransactionError):
    """Attempt to modify objects governed by a read-only data manager."""


class IObjectSerializer(zope.interface.Interface):
    """An object serializer allows for custom serialization output for
    objects."""

    def can_read(state):
        """Returns a boolean indicating whether this serializer can deserialize
        this state."""

    def get_object(state):
        """Convert the state to an object."""

    def can_write(obj):
        """Returns a boolean indicating whether this serializer can serialize
        this object."""

    def get_state(obj):
        """Convert the object to a state/document."""


class IObjectWriter(zope.interface.Interface):
    """The object writer stores an object in the database."""

    def get_non_persistent_state(obj, seen):
        """Convert a non-persistent object to a JSONB state/document."""

    def get_persistent_state(obj, seen):
        """Convert a persistent object to a JSONB state/document."""

    def get_state(obj, seen=None):
        """Convert an arbitrary object to a JSONB state/document.

        A ``CircularReferenceError`` is raised, if a non-persistent loop is
        detected.
        """

    def store(obj, id=None):
        """Store an object in the database with given id

        If id is not specified, unique one will be generated
        """


class IObjectReader(zope.interface.Interface):
    """The object reader reads an object from the database."""

    def resolve(path):
        """Resolve a path to a class.

        The path can be any string. It is the responsibility of the resolver
        to maintain the mapping from path to class.
        """

    def get_object(state, obj):
        """Get an object from the given state.

        The ``obj`` is the JSONB document of which the created object is part
        of.
        """

    def set_ghost_state(obj):
        """Convert a ghosted object to an active object by loading its state.
        """

    def get_ghost(coll_name, oid):
        """Get the ghosted version of the object.
        """


class IPJConnectionPool(zope.interface.Interface):
    def getconn():
        """Get connection from the pool"""

    def putconn(conn):
        """Put conncetion back to the pool"""

    def closeall():
        """Close all connection in the pool"""

    def reopen():
        """Reopen the pool after it was closed"""


class IPJDataManager(persistent.interfaces.IPersistentDataManager):
    """A persistent data manager that stores data in PostGreSQL/JSONB."""

    root = zope.interface.Attribute(
        """Get the root object, which is a mapping.""")

    def create_tables(tables):
        """Create passed tables and persistence_name_map, use this instead
        of PJ_AUTO_CREATE_TABLES"""

    def get_table_of_object(obj):
        """Return the table name for an object."""

    def reset():
        """Reset the datamanager for the next transaction."""

    def dump(obj):
        """Store the object to PostGreSQL/JSONB and return its DBRef."""

    def load(dbref):
        """Load the object from PostGreSQL/JSONB by using its DBRef.

        Note: The returned object is in the ghost state.
        """

    def flush():
        """Flush all changes to PostGreSQL."""

    def insert(obj, id=None):
        """Insert an object into PostGreSQL.

        The correct collection is determined by object type.

        If `id` is provided, object will be inserted under that id. Otherwise,
        a new unique id will be generated.
        """

    def remove(obj):
        """Remove an object from PostGreSQL.

        The correct collection is determined by object type.
        """

    def isDirty():
        """Return a bool whether there was any database write since the
        transaction started.

        A SQL write command or any persistent object change will set this flag.

        `transaction.commit` or `transaction.abort` clears this flag.
        """

    def setDirty():
        """Set dirty flag"""

    def setTransactionOptions(
        readonly: bool = None,
        deferrable: bool = None,
        isolation_level: int = None) -> None:
        """Set the options for the future transaction

        Options can only be set before the postgres transaction has started
        """


class IPJDataManagerProvider(zope.interface.Interface):
    """Utility to get a PJ data manager.

    Implementations of this utility usually maintain connection information
    and ensure that there is one consistent datamanager per thread.
    """

    def get(database):
        """Return a PJ data manager for the given database."""


class IPersistentSerializationHooks(zope.interface.Interface):
    """Persistent Serialization Hooks

    Persistent objects implementing this interface will be given the
    opportunity during storage proceedings to consuct some custom
    serialization tasks.
    """

    def _pj_after_store_hook(conn):
        """Custom store hook.

        This method is called after the object has been stored in the
        database. The ``conn`` attribute is a connection to the database.

        You should not manage the connection within this method, ie. abort or
        commit the transaction.
        """

    def _pj_after_load_hook(conn):
        """Custom load hook.

        This method is called after the object has been loaded fully from the
        database. The ``conn`` attribute is a connection to the database.

        You should not manage the connection within this method, ie. abort or
        commit the transaction.
        """


class IColumnSerialization(zope.interface.Interface):
    """Column Serialization Support

    Persistent objects implementing this interfave are able to store some of
    their data into columns of their storage table. This allows developers to
    access those attributes in a more natural way using SQL.
    """

    _pj_column_fields = zope.schema.Tuple(
        title='Column Fields',
        description=(
            'A list of zope.schema fields that represent columns '
            'in the storage table. Fields cannot be named `id` or '
            '`data` as those attributes are reserved.'
            ''
            'The fields are used to '
            '- auto create the SQL table '
            '- populate native SQL column values by `_pj_get_column_fields` '
            'see also pjpersist.zope.container.PJContainer._pj_column_fields'
            'that list of field names should be in sync with this'),
        required=True)

    def _pj_get_column_fields():
        """Get Column Fields values as a mapping from name to value."""



##############################################################################
#
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Persistent Object Support"""
import persistent
import zope.interface

from pjpersist import interfaces


@zope.interface.implementer(interfaces.IPersistentSerializationHooks)
class PersistentSerializationHooks(persistent.Persistent):

    def _pj_after_store_hook(self, conn):
        return None

    def _pj_after_load_hook(self, conn):
        return None


@zope.interface.implementer(interfaces.IColumnSerialization)
class SimpleColumnSerialization(object):

    _pj_column_fields = ()

    def _pj_get_column_fields(self):
        return {
            field.__name__: getattr(self, field.__name__)
            for field in self._pj_column_fields}


# XXX: extend later to a z3c.form'ish Fields to have `omit` too
def select_fields(schema, *fieldnames):
    return tuple(schema[fname] for fname in fieldnames)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import datetime
import doctest
import json
from pprint import pprint

from pjpersist import datamanager
from pjpersist import testing
from pjpersist import serialize
from pjpersist import sqlbuilder as sb

DTIMES = [
    datetime.datetime(2013,1,22,18,59),
    datetime.datetime(1987,2,2,8,31),
    datetime.datetime(1949,6,25,23,50),
    datetime.datetime(1918,3,3,8,35),
    datetime.datetime(1980,7,23,8,6),
    datetime.datetime(2011,11,1,16,26),
    datetime.datetime(1911,1,17,4,24),
    datetime.datetime(1975,7,22,4,44),
    datetime.datetime(1914,5,17,9,37),
    None,
    datetime.datetime(1936,12,16,10,38),
    datetime.datetime(1930,12,12,3,58),
    datetime.datetime(1995,1,23,9,59),
    datetime.datetime(1923,6,2,15,5),
    datetime.datetime(1967,3,1,13,42),
    datetime.datetime(1966,7,22,17,30),
    datetime.datetime(2005,7,13,17,18),
    datetime.datetime(1912,1,9,15,4),
    datetime.datetime(1983,6,27,3,56),
    None,
]


def pjvalue(obj):
    return serialize.ObjectWriter(None).get_state(obj)


def jformat(obj):
    """Helper for inserting JSON"""
    return datamanager.Json(pjvalue(obj))


def setUp(test):
    testing.setUp(test)

    conn = test.globs['conn']
    with conn.cursor() as cur:
        cur.execute("CREATE TABLE ser (id SERIAL PRIMARY KEY, data JSONB)")
        for dt in DTIMES:
            obj = dict(dtime=dt)
            cur.execute("INSERT INTO ser (data) VALUES (%s)", (jformat(obj),))
    conn.commit()


def select(conn, query, print_sql=False, **kwargs):
    try:
        with conn.cursor() as cur:
            sql = sb.sqlrepr(
                sb.Select(sb.Field("ser", "data"), where=query, **kwargs),
                'postgres'
            )
            if print_sql:
                print('SQL> ', sql)
            cur.execute(sql)
            for e in cur.fetchall():
                pprint(e[0])
    finally:
        conn.rollback()


def doctest_datetime_range():
    """Test datetime serialization vs. SQL range query

    >>> datafld = sb.Field('ser', 'data')
    >>> select(conn, sb.JGET(datafld, 'dtime') == pjvalue(DTIMES[10]), True)
    SQL>  SELECT ser.data FROM ser WHERE (((ser.data) -> ('dtime')) = ('{"_py_type": "datetime.datetime", "value": "1936-12-16T10:38:00.000000"}'::jsonb))
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1936-12-16T10:38:00.000000'}}

    >>> select(conn, sb.JGET(datafld, 'dtime') > pjvalue(DTIMES[10]),
    ...     print_sql=True, orderBy="(data->'dtime')")
    SQL>  SELECT ser.data FROM ser WHERE (((ser.data) -> ('dtime')) > ('{"_py_type": "datetime.datetime", "value": "1936-12-16T10:38:00.000000"}'::jsonb)) ORDER BY (data->'dtime')
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1949-06-25T23:50:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1966-07-22T17:30:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1967-03-01T13:42:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1975-07-22T04:44:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1980-07-23T08:06:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1983-06-27T03:56:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1987-02-02T08:31:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1995-01-23T09:59:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '2005-07-13T17:18:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '2011-11-01T16:26:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '2013-01-22T18:59:00.000000'}}

    >>> sorted([dt for dt in DTIMES if dt > DTIMES[10]])
    Traceback (most recent call last):
    ...
    TypeError: ...

    >>> select(conn, sb.JGET(datafld, 'dtime') < pjvalue(DTIMES[10]),
    ...     print_sql=True, orderBy="(data->'dtime')")
    SQL>  SELECT ser.data FROM ser WHERE (((ser.data) -> ('dtime')) < ('{"_py_type": "datetime.datetime", "value": "1936-12-16T10:38:00.000000"}'::jsonb)) ORDER BY (data->'dtime')
    {'dtime': None}
    {'dtime': None}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1911-01-17T04:24:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1912-01-09T15:04:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1914-05-17T09:37:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1918-03-03T08:35:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1923-06-02T15:05:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1930-12-12T03:58:00.000000'}}

    >>> select(conn, sb.JGET(datafld, 'dtime') == None,
    ...     print_sql=True, orderBy="(data->'dtime')")
    SQL>  SELECT ser.data FROM ser WHERE (((ser.data) -> ('dtime')) = ('null'::jsonb)) ORDER BY (data->'dtime')
    {'dtime': None}
    {'dtime': None}

    >>> select(conn, sb.JGET(datafld, 'dtime') != None,
    ...     print_sql=True, orderBy="(data->'dtime')")
    SQL>  SELECT ser.data FROM ser WHERE (((ser.data) -> ('dtime')) <> ('null'::jsonb)) ORDER BY (data->'dtime')
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1911-01-17T04:24:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1912-01-09T15:04:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1914-05-17T09:37:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1918-03-03T08:35:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1923-06-02T15:05:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1930-12-12T03:58:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1936-12-16T10:38:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1949-06-25T23:50:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1966-07-22T17:30:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1967-03-01T13:42:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1975-07-22T04:44:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1980-07-23T08:06:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1983-06-27T03:56:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1987-02-02T08:31:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '1995-01-23T09:59:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '2005-07-13T17:18:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '2011-11-01T16:26:00.000000'}}
    {'dtime': {'_py_type': 'datetime.datetime',
                'value': '2013-01-22T18:59:00.000000'}}

    """


def test_suite():
    suite = doctest.DocTestSuite(
        setUp=setUp, tearDown=testing.tearDown,
        checker=testing.checker,
        optionflags=testing.OPTIONFLAGS)
    suite.layer = testing.db_layer
    return suite



import datetime
BIGDICT = {'Alcestis': 'mantilla',
 'arraignments': True,
 'bosh': 6909,
 'chant': 6785,
 'garbanzo': {'Armenian': True,
              'benign': {'easygoing': {'basted': 'counterpoint',
                                       'nonpolluting': 'Gretchen',
                                       'ointment': {'Talleyrand': False,
                                                    'comparably': 'hedgehog',
                                                    'goiter': 6895,
                                                    'medalist': {},
                                                    'pinked': False},
                                       'replication': datetime.datetime(1918, 3, 3, 8, 35),
                                       'ripsaw': 4316},
                         'interaction': datetime.datetime(1980, 7, 23, 8, 6),
                         'renaissances': datetime.datetime(2013, 1, 22, 18, 59),
                         'revision': True,
                         'substantiates': {'Pepsi': True,
                                           'affords': 5854,
                                           'buckskins': 'legalize',
                                           'syphilis': False,
                                           'underdeveloped': True}},
              'pennants': {'armadillo': {'Condillac': 'Equuleus',
                                         'Orly': datetime.datetime(1911, 1, 17, 4, 24),
                                         'civvies': 9109,
                                         'hilly': datetime.datetime(2013, 1, 22, 18, 59),
                                         'nurture': 1718},
                           'comers': 322,
                           'furtherance': {'Brahe': datetime.datetime(1980, 7, 23, 8, 6),
                                           'Whitehead': False,
                                           'joysticks': {'drinker': 6444,
                                                         'gamed': 'hosing',
                                                         'larkspur': 3430,
                                                         'prepaid': 'unspoiled',
                                                         'stockpile': datetime.datetime(1923, 6, 2, 15, 5)},
                                           'snarls': 'meditates',
                                           'studding': {'Botticelli': 8214,
                                                        'Uruguayans': 830,
                                                        'illicitness': 7397,
                                                        'rosin': False,
                                                        'underskirts': False}},
                           'immodestly': 4678,
                           'kleptomania': 'pounced'},
              'sachet': {'bankrolls': True,
                         'breeding': 9051,
                         'failure': datetime.datetime(1975, 7, 22, 4, 44),
                         'imposed': 'avenged',
                         'undeceived': {'festal': 3225,
                                        'intermezzi': 1484,
                                        'poor': False,
                                        'reprinted': datetime.datetime(1918, 3, 3, 8, 35),
                                        'sleepwalkers': 'embarrassing'}},
              'tryst': True},
 'inaugural': {'Horacio': datetime.datetime(1923, 6, 2, 15, 5),
               'Utopians': True,
               'chimeras': datetime.datetime(1967, 3, 1, 13, 42),
               'domed': 3535,
               'member': False},
 'lava': 'Domingo',
 'potash': 7166,
 'spoils': False,
 'subteens': False}

HUGEDICT = {'Agnes': 9671,
 'Alcuin': False,
 'Borges': [{'cokes': [],
             'contenting': {'blankly': {'Brownian': 2967,
                                        'bondage': 'Brant',
                                        'clubfoot': [],
                                        'hemophiliacs': 8448,
                                        'superstitious': datetime.datetime(1995, 1, 23, 9, 59)},
                            'gofers': datetime.datetime(1967, 3, 1, 13, 42),
                            'lifeblood': 'feisty',
                            'tomboy': 'filibusters',
                            'translated': []},
             'indifference': False,
             'mania': datetime.datetime(1987, 2, 2, 8, 31),
             'sweetly': True},
            'assessment'],
 'Braille': 'audio',
 'Cancer': 'reoccurs',
 'Devon': 'substituting',
 'Gilead': [],
 'Harry': 'seated',
 'Herschel': 'jack',
 'Horne': 6853,
 'Levine': datetime.datetime(1918, 3, 3, 8, 35),
 'Lugosi': {'accumulating': [False,
                             {'Tomsk': [True],
                              'forsakes': 7915,
                              'minuscules': 6354,
                              'petioles': False,
                              'waterside': 7134}],
            'depreciating': 3519,
            'fogey': 920,
            'outbalancing': 'thoughts',
            'porpoising': 3589},
 'Max': datetime.datetime(2005, 7, 13, 17, 18),
 'Mountbatten': 2791,
 'Oprah': False,
 'Peck': False,
 'Percival': [False,
              [datetime.datetime(1967, 3, 1, 13, 42),
               269,
               True,
               4595,
               {'Ovid': {'carpentered': 3486,
                         'chessman': [],
                         'exertion': 6442,
                         'masculine': 5995,
                         'unscrewed': 'otiose'},
                'Thaddeus': 'Botswana',
                'acknowledging': 5445,
                'dipstick': 'shimmered',
                'involuntarily': True}]],
 'Rhodes': 'spliced',
 'Rick': 3390,
 'Saskatoon': [],
 'Sudetenland': {'Brinkley': 2652,
                 'Nestle': {'agilely': 6234,
                            'execrate': datetime.datetime(2005, 7, 13, 17, 18),
                            'fittingly': datetime.datetime(1923, 6, 2, 15, 5),
                            'infomercials': datetime.datetime(1912, 1, 9, 15, 4),
                            'sarcastically': {'Verona': {'functions': False,
                                                         'hastiest': 8998,
                                                         'push': datetime.datetime(1936, 12, 16, 10, 38),
                                                         'sailors': {},
                                                         'smother': 'downpours'},
                                              'fiendishly': 6242,
                                              'gallstones': [[],
                                                             datetime.datetime(1967, 3, 1, 13, 42),
                                                             {},
                                                             False],
                                              'riles': datetime.datetime(1949, 6, 25, 23, 50),
                                              'vegetable': 'nursemaid'}},
                 'dauntlessly': [True,
                                 [],
                                 {'cask': 6334,
                                  'gamuts': {'Java': 8903,
                                             'Vilma': 7984,
                                             'god': {},
                                             'provisions': [],
                                             'shredded': []},
                                  'housings': 'Australasia',
                                  'stooping': [True,
                                               'stirrers',
                                               {},
                                               False,
                                               {}],
                                  'tied': True}],
                 'gruffly': 1466,
                 'roaches': 5433},
 'Veda': 'enterprise',
 'Vesalius': datetime.datetime(2011, 11, 1, 16, 26),
 'Vietnam': 3655,
 'Wilberforce': False,
 'Wolfe': False,
 'Wuhan': 'shippers',
 'Zeke': {'abusive': {'betoken': 1922,
                      'eigenvalues': {'Nebraskan': None,
                                      'chortles': datetime.datetime(1936, 12, 16, 10, 38),
                                      'engrossed': True,
                                      'harridan': {'Roquefort': [],
                                                   'canvased': [],
                                                   'reproduce': datetime.datetime(2005, 7, 13, 17, 18),
                                                   'spacesuit': False,
                                                   'watermarking': datetime.datetime(1918, 3, 3, 8, 35)},
                                      'recommences': None},
                      'fevers': datetime.datetime(1914, 5, 17, 9, 37),
                      'miff': [[2624,
                                datetime.datetime(1995, 1, 23, 9, 59),
                                430],
                               {'Hay': [],
                                'absolving': {},
                                'bibliography': [],
                                'cantons': datetime.datetime(2005, 7, 13, 17, 18),
                                'necking': datetime.datetime(1975, 7, 22, 4, 44)},
                               [[], {}],
                               'tincturing',
                               6372],
                      'subjugates': 3413},
          'basely': datetime.datetime(1949, 6, 25, 23, 50),
          'chiselers': {'Memphis': [False,
                                    True,
                                    [datetime.datetime(1918, 3, 3, 8, 35),
                                     {}]],
                        'incessantly': {'excrement': datetime.datetime(1923, 6, 2, 15, 5),
                                        'opaquing': 'deviltry',
                                        'patriotically': datetime.datetime(1936, 12, 16, 10, 38),
                                        'robbing': False,
                                        'stanching': False},
                        'nightgown': datetime.datetime(1912, 1, 9, 15, 4),
                        'papyrus': {'Malachi': False,
                                    'ditch': False,
                                    'flypapers': datetime.datetime(2011, 11, 1, 16, 26),
                                    'nous': None,
                                    'swirled': datetime.datetime(1980, 7, 23, 8, 6)},
                        'shirred': False},
          'jot': 'wrestles',
          'squeezer': ['orbs']},
 'Zomba': None,
 'abducts': {'Pei': datetime.datetime(1995, 1, 23, 9, 59),
             'floorboard': 'curvy',
             'gulling': 'inhaling',
             'pungently': True,
             'vodka': 'heifers'},
 'abortionists': datetime.datetime(2011, 11, 1, 16, 26),
 'accomplishing': True,
 'accounting': [7812, 6998],
 'adapter': {'brainwashed': [datetime.datetime(1967, 3, 1, 13, 42),
                             None,
                             'confrontations'],
             'earthquake': [{'bus': [5385, 'unburdening', [], {}],
                             'caustic': 'peritonitis',
                             'eulogies': 9254,
                             'intimation': [{},
                                            datetime.datetime(2005, 7, 13, 17, 18),
                                            datetime.datetime(1936, 12, 16, 10, 38),
                                            False],
                             'shakiest': ['permanence', 'madrigal']},
                            6064],
             'indomitably': False,
             'liar': ['mesmerized'],
             'sectioning': [['Parsi', datetime.datetime(1983, 6, 27, 3, 56)],
                            datetime.datetime(1949, 6, 25, 23, 50),
                            datetime.datetime(1980, 7, 23, 8, 6),
                            5570,
                            datetime.datetime(1923, 6, 2, 15, 5)]},
 'aesthetes': {'berths': True,
               'massages': False,
               'method': datetime.datetime(1918, 3, 3, 8, 35),
               'pealing': 4645,
               'synod': []},
 'alighted': False,
 'angst': datetime.datetime(1912, 1, 9, 15, 4),
 'apprenticeship': [3713,
                    {'braked': 6364,
                     'chimpanzees': datetime.datetime(1995, 1, 23, 9, 59),
                     'floggings': 'yearnings',
                     'notoriety': datetime.datetime(1980, 7, 23, 8, 6),
                     'tranquilest': [6857]},
                    datetime.datetime(1975, 7, 22, 4, 44)],
 'arcing': 'Holmes',
 'ashore': False,
 'backward': 6032,
 'behoove': 4094,
 'bigmouths': 'scrapers',
 'blinded': 'nonresident',
 'bombshell': 1252,
 'both': 'arrow',
 'bouncy': None,
 'bunked': datetime.datetime(1995, 1, 23, 9, 59),
 'buy': [datetime.datetime(1995, 1, 23, 9, 59)],
 'canonical': {'backfields': 'cosignatories',
               'fain': datetime.datetime(1983, 6, 27, 3, 56),
               'frighting': datetime.datetime(1930, 12, 12, 3, 58),
               'shimmy': 5768,
               'spastics': datetime.datetime(1911, 1, 17, 4, 24)},
 'canonizations': 'battle',
 'catalogs': False,
 'categorization': False,
 'charade': 7048,
 'cholesterol': datetime.datetime(1983, 6, 27, 3, 56),
 'collected': [],
 'communicated': [datetime.datetime(1975, 7, 22, 4, 44),
                  True,
                  7047,
                  {'aspirating': 4305,
                   'conjoining': 'privy',
                   'damask': 'impala',
                   'hearts': [],
                   'unworthy': 7095}],
 'competences': 2791,
 'confederacy': 'polliwogs',
 'conjecture': ['salubrious', 5367, 9920],
 'contaminant': datetime.datetime(1911, 1, 17, 4, 24),
 'contravention': datetime.datetime(1949, 6, 25, 23, 50),
 'coping': [9861, False, 3763],
 'cornerstones': {'evildoers': ['chowder', 4864],
                  'passageway': datetime.datetime(2011, 11, 1, 16, 26),
                  'pastry': {'fetishists': [687, None, 5846],
                             'giblets': [6903,
                                         [[],
                                          [],
                                          'corporals',
                                          'glinted',
                                          {}],
                                         [datetime.datetime(1936, 12, 16, 10, 38),
                                          1666,
                                          True,
                                          {},
                                          False],
                                         5831,
                                         8925],
                             'glorifying': 'staked',
                             'maestro': 'amalgam',
                             'villainy': 'ungodlier'},
                  'perfidious': [5013,
                                 {'aromas': None,
                                  'boilerplate': {'foreseeable': datetime.datetime(2011, 11, 1, 16, 26),
                                                  'kidnapping': True,
                                                  'monotheistic': 'repeals',
                                                  'pejoratives': [],
                                                  'thwacked': 'Nicaea'},
                                  'casseroles': True,
                                  'indicted': {'Gaines': 'liaises',
                                               'manse': [],
                                               'marauded': 'Praetorian',
                                               'prefix': datetime.datetime(1936, 12, 16, 10, 38),
                                               'sweetbriar': 'Barbuda'},
                                  'transgresses': [False]},
                                 [datetime.datetime(1987, 2, 2, 8, 31),
                                  6653,
                                  {'Dachau': [],
                                   'anger': datetime.datetime(1918, 3, 3, 8, 35),
                                   'expatiate': True,
                                   'portioning': False,
                                   'swooning': 8374},
                                  [],
                                  6871],
                                 8436],
                  'retirement': 7577},
 'corollaries': {'Ger': datetime.datetime(1923, 6, 2, 15, 5),
                 'downloads': 9182,
                 'hospitable': datetime.datetime(1949, 6, 25, 23, 50),
                 'playfulness': 5617,
                 'urethra': datetime.datetime(1914, 5, 17, 9, 37)},
 'corrections': 8263,
 'courtliness': 'poorer',
 'cozy': False,
 'cuddlier': [datetime.datetime(1966, 7, 22, 17, 30), 'tutored'],
 'cumming': 'ghostlier',
 'darnedest': False,
 'declaims': {'analyticalally': datetime.datetime(1980, 7, 23, 8, 6),
              'availing': {'calumniated': datetime.datetime(2005, 7, 13, 17, 18),
                           'chaffs': {'Cadillac': [{}, 'Tucker', {}],
                                      'Pecos': 'Sarto',
                                      'caw': {'disrespectfully': [],
                                              'maypole': [],
                                              'stagnant': {},
                                              'unite': {},
                                              'zap': {}},
                                      'javelin': [{}, 8780],
                                      'sycamore': 7456},
                           'enchanted': False,
                           'manholes': False,
                           'miss': [{'Trobriand': {},
                                     'dialects': 'preoccupation',
                                     'elide': datetime.datetime(1980, 7, 23, 8, 6),
                                     'fabricate': {},
                                     'smackers': datetime.datetime(1967, 3, 1, 13, 42)},
                                    datetime.datetime(1983, 6, 27, 3, 56),
                                    5817,
                                    [True,
                                     datetime.datetime(1918, 3, 3, 8, 35),
                                     {},
                                     'postponement',
                                     8970],
                                    False]},
              'crackpot': 2879,
              'shufflers': [{'besieged': {'fixity': 9979,
                                          'heave': datetime.datetime(1912, 1, 9, 15, 4),
                                          'overseers': datetime.datetime(1936, 12, 16, 10, 38),
                                          'plowmen': True,
                                          'triptychs': 1072},
                             'central': 5135,
                             'faultier': 'sweetens',
                             'maturer': ['opuses', []],
                             'proverbially': 2744},
                            True],
              'spoon': 4998},
 'decomposes': [False],
 'dehumanize': True,
 'depopulating': 'Vulgate',
 'deride': datetime.datetime(1975, 7, 22, 4, 44),
 'destructs': 'unfathomable',
 'detonators': True,
 'disabused': {'Dumpster': {'Iva': ['freeloaded',
                                    True,
                                    ['misplayed',
                                     {},
                                     [],
                                     False,
                                     datetime.datetime(1918, 3, 3, 8, 35)],
                                    5169,
                                    False],
                            'faker': 'gunning',
                            'scorecard': {'Brittany': 5579,
                                          'maltreat': False,
                                          'promptings': {'furze': 'Karenina',
                                                         'importuning': {},
                                                         'litigiousness': 'tenderfoots',
                                                         'sensitives': 'reassuringly',
                                                         'welshing': 3201},
                                          'selectors': datetime.datetime(2013, 1, 22, 18, 59),
                                          'spray': datetime.datetime(1914, 5, 17, 9, 37)},
                            'sleepwalking': {'asphalting': 6997,
                                             'dish': [],
                                             'drain': 'surfeiting',
                                             'fleeces': {'admired': 2124,
                                                         'disappoint': 'earthly',
                                                         'farina': False,
                                                         'klutzier': 'nightfall',
                                                         'sublimation': {}},
                                             'sleepier': True},
                            'split': {'Rather': [{},
                                                 'transference',
                                                 True,
                                                 datetime.datetime(1914, 5, 17, 9, 37)],
                                      'Ymir': False,
                                      'antenna': True,
                                      'basal': {'Brahmin': {},
                                                'beneficent': datetime.datetime(1936, 12, 16, 10, 38),
                                                'depositions': datetime.datetime(1949, 6, 25, 23, 50),
                                                'scums': {},
                                                'uphills': None},
                                      'commandoes': 'lipreading'}},
               'Glaswegian': 'juicier',
               'collaborate': False,
               'recessed': {'astronaut': datetime.datetime(1975, 7, 22, 4, 44),
                            'laughter': [datetime.datetime(1983, 6, 27, 3, 56),
                                         8038,
                                         True,
                                         {'Jose': [],
                                          'ballsier': {},
                                          'casks': {},
                                          'harnessing': 'artificial',
                                          'integrated': 'alphabet'},
                                         True],
                            'ornate': {'Mondrian': {'Corfu': True,
                                                    'blithely': {},
                                                    'essence': 7905,
                                                    'lithe': datetime.datetime(1930, 12, 12, 3, 58),
                                                    'obscurity': {}},
                                       'ayes': datetime.datetime(1912, 1, 9, 15, 4),
                                       'homeopathy': True,
                                       'neglect': {'Abrams': False,
                                                   'gook': True,
                                                   'legitimizes': False,
                                                   'pervades': {},
                                                   'tillable': datetime.datetime(1923, 6, 2, 15, 5)},
                                       'slaver': [521]},
                            'shaken': 3158,
                            'workers': 3137},
               'silvery': datetime.datetime(1923, 6, 2, 15, 5)},
 'disco': True,
 'discrimination': True,
 'displeases': [],
 'doyens': [[2614, False, 9485, datetime.datetime(1983, 6, 27, 3, 56)],
            {'Church': datetime.datetime(1923, 6, 2, 15, 5),
             'Devin': {'Hispanic': [558, [], []],
                       'asylums': 'denotes',
                       'brusquer': {'battlefields': 9566,
                                    'drape': 1006,
                                    'flagellate': 'rued',
                                    'logic': True,
                                    'randomized': {}},
                       'incursions': 'briefcases',
                       'initiates': 'Brahmanisms'},
             'confided': {'approbation': 7365,
                          'duded': 'visualizing',
                          'enmesh': True,
                          'separately': datetime.datetime(1967, 3, 1, 13, 42),
                          'xylophone': 7646},
             'mugged': False,
             'reminiscence': True},
            [{'Cohan': False,
              'Mesmer': datetime.datetime(1966, 7, 22, 17, 30),
              'coke': 803,
              'cozily': True,
              'extort': 'entirely'},
             [True, [2782, 8506], 'scintillation', 3006],
             5032,
             {'Grumman': datetime.datetime(1936, 12, 16, 10, 38),
              'cloaks': 'adverbial',
              'culprit': 8740,
              'egocentrics': datetime.datetime(1980, 7, 23, 8, 6),
              'objectionable': False}],
            {'Belgrade': [],
             'Lagos': True,
             'agribusinesses': 'mercerize',
             'enamor': False,
             'exigency': 'psychogenic'},
            {'divinely': {'coiled': True,
                          'jinns': [{}],
                          'multitudes': [[],
                                         datetime.datetime(2005, 7, 13, 17, 18),
                                         [],
                                         []],
                          'reactionaries': 8034,
                          'stairwells': {'gasping': datetime.datetime(1936, 12, 16, 10, 38),
                                         'grouted': datetime.datetime(1914, 5, 17, 9, 37),
                                         'paraffin': {},
                                         'redheaded': datetime.datetime(1949, 6, 25, 23, 50),
                                         'remakes': datetime.datetime(1967, 3, 1, 13, 42)}},
             'easily': 'backlogged',
             'lauds': 3708,
             'philharmonic': {'butterfingers': 'Jordan',
                              'cleavers': 'leash',
                              'gloamings': datetime.datetime(1975, 7, 22, 4, 44),
                              'lament': {'Moldavia': [],
                                         'Moore': {},
                                         'newsboys': True,
                                         'parapets': 'revolutionary',
                                         'savored': False},
                              'surveyor': 3300},
             'wormiest': {'draconian': 'evillest',
                          'quoiting': 1886,
                          'restores': {'constricted': datetime.datetime(1980, 7, 23, 8, 6),
                                       'enjoyed': 'remains',
                                       'grammars': datetime.datetime(1987, 2, 2, 8, 31),
                                       'hateful': 'brotherliness',
                                       'temporary': False},
                          'sputters': [[],
                                       [],
                                       True,
                                       'lipsticks',
                                       datetime.datetime(1923, 6, 2, 15, 5)],
                          'valedictory': 'correlated'}}],
 'dumfounding': datetime.datetime(1975, 7, 22, 4, 44),
 'earthly': 'colonial',
 'edibles': 'enraging',
 'encored': datetime.datetime(1930, 12, 12, 3, 58),
 'extraordinarily': 'Kirsten',
 'eyepiece': True,
 'facsimiled': 'hermetic',
 'fanzine': [[{'abnegated': 'unholiest',
               'banning': 'indicatives',
               'cameras': False,
               'insecticide': [{}, [], 2702],
               'yielded': {'bloodhounds': 4940,
                           'gendarmes': [],
                           'isinglass': [],
                           'ridging': False,
                           'stripling': []}},
              False,
              4290,
              3428],
             True,
             datetime.datetime(1966, 7, 22, 17, 30),
             8379,
             'grain'],
 'filching': [datetime.datetime(1930, 12, 12, 3, 58),
              False,
              2491,
              {'brutalize': {'Polly': False,
                             'Toledo': 3889,
                             'gladiolus': 'rundowns',
                             'heartiest': 6050,
                             'spite': {'Pygmies': {},
                                       'axiomatic': {},
                                       'degrees': 347,
                                       'dismember': 'mayhem',
                                       'retrospect': {}}},
               'docket': datetime.datetime(1936, 12, 16, 10, 38),
               'purplish': datetime.datetime(1967, 3, 1, 13, 42),
               'scientists': [],
               'spates': 'patriarch'},
              {'Manet': 5895,
               'bashed': 9585,
               'coveralls': {'frown': 2119,
                             'heathens': False,
                             'hovel': 'gnaws',
                             'slovenliness': [],
                             'unorthodox': ['riming',
                                            'mortgager',
                                            'encamps',
                                            False,
                                            []]},
               'envisioning': False,
               'undo': datetime.datetime(1980, 7, 23, 8, 6)}],
 'fillips': datetime.datetime(1914, 5, 17, 9, 37),
 'filter': datetime.datetime(1966, 7, 22, 17, 30),
 'fleeting': [[['proceeds'], 'northwestern'], 'gray'],
 'flicking': True,
 'floodlit': True,
 'flunking': True,
 'forged': 8401,
 'fryer': datetime.datetime(1975, 7, 22, 4, 44),
 'garbageman': 6151,
 'gigglier': {'Reinaldo': {'breached': datetime.datetime(1930, 12, 12, 3, 58),
                           'kidney': 7671,
                           'likely': datetime.datetime(2005, 7, 13, 17, 18),
                           'orphans': {'Handel': {'Mitch': 'bowdlerized',
                                                  'Okefenokee': 9560,
                                                  'auto': True,
                                                  'hurtful': [],
                                                  'rediscovery': 'deplanes'},
                                       'creativity': {'Ouagadougou': 5593,
                                                      'Stirling': [],
                                                      'filet': None,
                                                      'posts': False,
                                                      'stranglehold': {}},
                                       'feral': 'Bastille',
                                       'gerrymandering': 'Udall',
                                       'xenon': datetime.datetime(1911, 1, 17, 4, 24)},
                           'yawed': True},
              'clerk': [[]],
              'fetuses': 2858,
              'happen': True,
              'sappy': datetime.datetime(1966, 7, 22, 17, 30)},
 'goggling': 7184,
 'groupie': {'Marcia': {'baboons': 4365,
                        'jobless': [datetime.datetime(2005, 7, 13, 17, 18),
                                    'implacable',
                                    {'Janjaweed': [],
                                     'Morita': 7946,
                                     'bouillabaisses': False,
                                     'geeing': datetime.datetime(1995, 1, 23, 9, 59),
                                     'overwrites': []}],
                        'propound': 976,
                        'travestied': True,
                        'vamped': datetime.datetime(1995, 1, 23, 9, 59)},
             'promos': ['curdling',
                        [datetime.datetime(1930, 12, 12, 3, 58),
                         datetime.datetime(1914, 5, 17, 9, 37),
                         {'conspirators': datetime.datetime(1923, 6, 2, 15, 5),
                          'flummoxes': 'hoots',
                          'masthead': 'desist',
                          'segregates': 'encumbered',
                          'weavers': 'glades'},
                         datetime.datetime(1914, 5, 17, 9, 37),
                         [3120]],
                        1007,
                        datetime.datetime(1987, 2, 2, 8, 31)],
             'slingshot': [],
             'spelling': datetime.datetime(1912, 1, 9, 15, 4),
             'tedious': {'Albee': 'vamped',
                         'Byelorussia': datetime.datetime(1914, 5, 17, 9, 37),
                         'fishtailed': False,
                         'lagging': True,
                         'megalopolis': datetime.datetime(1911, 1, 17, 4, 24)}},
 'hectoring': datetime.datetime(1987, 2, 2, 8, 31),
 'holdouts': ['pinions',
              {'Schiller': True,
               'dairyman': {'batty': 'ironies',
                            'enthrones': 'slicking',
                            'implementation': {'Fredrick': [],
                                               'Irma': True,
                                               'Kurile': datetime.datetime(1911, 1, 17, 4, 24),
                                               'daintiness': datetime.datetime(2013, 1, 22, 18, 59),
                                               'invaluable': 'woodpile'},
                            'obelisks': 3312,
                            'sleigh': datetime.datetime(1930, 12, 12, 3, 58)},
               'mainstreamed': [3505],
               'merry': datetime.datetime(1983, 6, 27, 3, 56),
               'reek': True},
              False,
              [{'Muawiya': 8138,
                'dosed': 'nonprescription',
                'ports': 'immunizing',
                'streptomycin': True,
                'trusty': {'Cornell': 'hushes',
                           'allergists': datetime.datetime(1930, 12, 12, 3, 58),
                           'clinging': {},
                           'mainsprings': 'bequeathed',
                           'tunnelled': datetime.datetime(1911, 1, 17, 4, 24)}},
               {'crimsons': datetime.datetime(1949, 6, 25, 23, 50),
                'grousing': datetime.datetime(1967, 3, 1, 13, 42),
                'nominee': 'overdosing',
                'pants': {'anecdote': False,
                          'cuttings': 'seeded',
                          'dolphin': 'purblind',
                          'recessives': 7595,
                          'satirist': 2061},
                'shinned': datetime.datetime(2013, 1, 22, 18, 59)},
               'cumbersome',
               {'chartreuse': True,
                'demoralizing': 'centipedes',
                'keys': {'Bahamas': 'soprano',
                         'Cepheus': [],
                         'hurriedly': 'deferentially',
                         'snowfall': 1276,
                         'unconventional': True},
                'kittens': 2653,
                'waxed': 'enthrals'},
               True],
              [{'aligned': True,
                'articulated': [{}, 6482, 'hybrid'],
                'canticles': 2660,
                'evening': [],
                'villainies': 1006},
               3786,
               [8586,
                'pioneers',
                {'abortionists': True,
                 'epithets': 'Fran',
                 'gangly': [],
                 'largess': [],
                 'pretending': 3879}],
               1473]],
 'horseradish': 'joy',
 'hostilities': {'Casals': {'clipt': datetime.datetime(1966, 7, 22, 17, 30),
                            'libations': 1207,
                            'servomechanisms': False,
                            'stopcock': 'digressed',
                            'usherette': True},
                 'Lenten': 2397,
                 'conditionally': True,
                 'meritocracies': {'attic': [],
                                   'eclipses': datetime.datetime(1987, 2, 2, 8, 31),
                                   'however': 'Mt',
                                   'manages': 'whirs',
                                   'nights': [[datetime.datetime(2005, 7, 13, 17, 18)],
                                              True,
                                              8485]},
                 'tuba': 3140},
 'humbler': ['soundproofs'],
 'humps': [3576,
           False,
           [[],
            datetime.datetime(1987, 2, 2, 8, 31),
            6898,
            datetime.datetime(1911, 1, 17, 4, 24),
            datetime.datetime(1995, 1, 23, 9, 59)]],
 'huskiness': 150,
 'imbalanced': 5365,
 'imperturbable': 1817,
 'incrusting': 3456,
 'intensifier': datetime.datetime(1966, 7, 22, 17, 30),
 'irritatingly': False,
 'itineraries': datetime.datetime(1912, 1, 9, 15, 4),
 'jeremiad': [5655],
 'judgements': datetime.datetime(1966, 7, 22, 17, 30),
 'knife': 2500,
 'lactates': {'amounting': 'interconnects',
              'awkwardest': True,
              'cistern': 3703,
              'gearbox': datetime.datetime(2013, 1, 22, 18, 59),
              'pastured': None},
 'laundering': datetime.datetime(2005, 7, 13, 17, 18),
 'litanies': 'Schroeder',
 'lumpiness': datetime.datetime(1983, 6, 27, 3, 56),
 'malpractice': False,
 'malt': [datetime.datetime(1995, 1, 23, 9, 59)],
 'menopausal': 'distrust',
 'middleman': {'Jerrold': 8321,
               'diddling': 'completer',
               'exegesis': False,
               'intransitives': 'sootier',
               'securing': False},
 'mil': {'chopstick': 'tingled',
         'controversy': datetime.datetime(1911, 1, 17, 4, 24),
         'gees': datetime.datetime(1911, 1, 17, 4, 24),
         'realest': 5065,
         'tyrants': True},
 'miscued': 2061,
 'misfires': datetime.datetime(1995, 1, 23, 9, 59),
 'modeling': [255,
              {'officers': {'afflicted': True,
                            'bamboozle': {'Cabot': {},
                                          'Li': 'thunderclouds',
                                          'grubbing': datetime.datetime(2013, 1, 22, 18, 59),
                                          'lowly': [],
                                          'reminisces': datetime.datetime(1918, 3, 3, 8, 35)},
                            'dialog': [True, [], {}, 2935],
                            'harmonization': datetime.datetime(1923, 6, 2, 15, 5),
                            'octettes': True},
               'scoops': 2437,
               'socializes': 149,
               'tropical': True,
               'vs': [6921, 7216, 'antiphonals', [{}, []]]}],
 'moonlighted': 2466,
 'mushes': 'Ikea',
 'muskrats': 'trinkets',
 'necked': 'aggrieve',
 'nominally': datetime.datetime(1912, 1, 9, 15, 4),
 'nullify': datetime.datetime(1923, 6, 2, 15, 5),
 'occult': 'molders',
 'oculars': {'Fridays': [False,
                         'telephone',
                         datetime.datetime(2005, 7, 13, 17, 18),
                         'creditable',
                         'smothered'],
             'Google': datetime.datetime(1995, 1, 23, 9, 59),
             'broader': [False,
                         True,
                         {'abated': False,
                          'biscuits': [datetime.datetime(2011, 11, 1, 16, 26),
                                       'slipknots'],
                          'nigh': 'Angolan',
                          'precursor': datetime.datetime(1912, 1, 9, 15, 4),
                          'tiptoeing': datetime.datetime(1967, 3, 1, 13, 42)},
                         [False, False]],
             'linkups': {'chessboards': [{'Omahas': True,
                                          'exclusively': False,
                                          'index': [],
                                          'misgovern': {},
                                          'unscrambles': False},
                                         'undue'],
                         'dazzled': datetime.datetime(1995, 1, 23, 9, 59),
                         'hellholes': 4717,
                         'tinted': 'erstwhile',
                         'travesties': 5393},
             'masts': False},
 'ordains': 'Caitlin',
 'oversimplifies': 'healthy',
 'parcelled': [],
 'patsies': [[],
             {'afforest': [],
              'incident': {'Virgil': datetime.datetime(1912, 1, 9, 15, 4),
                           'contents': 'row',
                           'dashiki': datetime.datetime(1912, 1, 9, 15, 4),
                           'gonorrhea': 8686,
                           'screeches': True},
              'overwhelm': datetime.datetime(2011, 11, 1, 16, 26),
              'parliamentarian': {'astrophysics': 5675,
                                  'enquired': True,
                                  'outsizes': {'Huff': datetime.datetime(1911, 1, 17, 4, 24),
                                               'crews': datetime.datetime(1930, 12, 12, 3, 58),
                                               'synonym': datetime.datetime(1983, 6, 27, 3, 56),
                                               'wangles': 4132,
                                               'weighs': {}},
                                  'whatsoever': [{},
                                                 'paranoids',
                                                 datetime.datetime(2011, 11, 1, 16, 26)],
                                  'yodellers': [197,
                                                [],
                                                False,
                                                datetime.datetime(1987, 2, 2, 8, 31)]},
              'triptychs': False},
             [],
             'Novembers'],
 'pigsties': ['delectation',
              {'arresting': False,
               'inner': 3268,
               'medications': {'beefs': 'prism',
                               'judgmental': [{}, {}],
                               'moguls': {'Brecht': 4463,
                                          'absorb': [],
                                          'trims': {},
                                          'tuns': {},
                                          'underestimating': {}},
                               'riles': 2493,
                               'shuttlecocked': True},
               'posthumously': 'opulent',
               'shoot': False},
              5265,
              {'Mirfak': 'gouger',
               'amaranths': datetime.datetime(1995, 1, 23, 9, 59),
               'email': datetime.datetime(2013, 1, 22, 18, 59),
               'engorge': 6681,
               'fetid': datetime.datetime(1987, 2, 2, 8, 31)},
              5694],
 'piking': False,
 'plants': True,
 'players': False,
 'plod': datetime.datetime(1918, 3, 3, 8, 35),
 'pointless': {'Lot': 'Cezanne',
               'bonnets': True,
               'glider': datetime.datetime(1949, 6, 25, 23, 50),
               'playback': {'birdbath': 1180,
                            'cabarets': 2531,
                            'oversupplied': 'viscountesses',
                            'protoplasm': 7240,
                            'wuss': 'wryly'},
               'stallion': {'pans': datetime.datetime(2005, 7, 13, 17, 18),
                            'photosynthesis': False,
                            'presidential': True,
                            'sufficiency': 'liqueurs',
                            'traders': datetime.datetime(1918, 3, 3, 8, 35)}},
 'polarization': 4126,
 'polliwog': ['humanism',
              7118,
              [datetime.datetime(2005, 7, 13, 17, 18), 6143]],
 'possession': 8822,
 'quash': [datetime.datetime(1987, 2, 2, 8, 31), 6777],
 'razzed': None,
 'reactivates': [datetime.datetime(1983, 6, 27, 3, 56)],
 'realities': 'refundable',
 'regimentation': True,
 'riddles': False,
 'romanced': 'extinguishers',
 'salesgirl': 3437,
 'saut\xc3\xa9ing': 'preoccupation',
 'scone': 8713,
 'senator': 'interpreted',
 'skepticism': {'Dec': 6776,
                'antiquary': ['setting',
                              datetime.datetime(1936, 12, 16, 10, 38),
                              {'Lemaitre': [8874, 9306],
                               'accentuate': {'flashlight': datetime.datetime(1966, 7, 22, 17, 30),
                                              'flooding': datetime.datetime(1936, 12, 16, 10, 38),
                                              'ladies': [],
                                              'ravening': [],
                                              'versification': {}},
                               'movables': datetime.datetime(1995, 1, 23, 9, 59),
                               'murmur': datetime.datetime(1949, 6, 25, 23, 50),
                               'parenthood': ['submerging', False]}],
                'grandparent': {'Vilma': {'Gallup': {'altimeter': 'overrode',
                                                     'burnouses': [],
                                                     'interloper': False,
                                                     'partitioning': datetime.datetime(1966, 7, 22, 17, 30),
                                                     'ukeleles': {}},
                                          'Gamay': False,
                                          'dashboard': 'wises',
                                          'mastoid': True,
                                          'terrible': datetime.datetime(1987, 2, 2, 8, 31)},
                                'agilest': datetime.datetime(1936, 12, 16, 10, 38),
                                'farrows': [False, 'Euterpe'],
                                'principals': {'bolts': datetime.datetime(1936, 12, 16, 10, 38),
                                               'leverages': 327,
                                               'rescuers': True,
                                               'scurrilous': {'castrating': False,
                                                              'charwomen': 6469,
                                                              'disuniting': {},
                                                              'ruffians': 'bunnies',
                                                              'statehood': datetime.datetime(1980, 7, 23, 8, 6)},
                                               'welching': 'disengagements'},
                                'sheepishness': [[1155, [], {}], 9034]},
                'habitu\xc3\xa9': {'Blacks': 'thymi',
                                   'Hans': False,
                                   'luminous': datetime.datetime(1914, 5, 17, 9, 37),
                                   'salubrious': 'contusions',
                                   'sonorous': 4393},
                'lucidness': {'Eyck': datetime.datetime(1949, 6, 25, 23, 50),
                              'bulging': [],
                              'feed': 'salmonellae',
                              'garrotes': datetime.datetime(1983, 6, 27, 3, 56),
                              'untested': None}},
 'slogging': datetime.datetime(2005, 7, 13, 17, 18),
 'smelliest': 8820,
 'smoker': 8780,
 'souvenir': ['regurgitation'],
 'sovereigns': 2534,
 'stormed': True,
 'stultified': [5535,
                'special',
                {'Berkshires': {'Nam': False,
                                'Weldon': {'Zachery': False,
                                           'betake': 6538,
                                           'cattleman': [],
                                           'milquetoasts': {},
                                           'undoubted': True},
                                'concreting': 'killed',
                                'cramps': [[],
                                           datetime.datetime(1912, 1, 9, 15, 4),
                                           {},
                                           False,
                                           {}],
                                'encumbrance': datetime.datetime(1966, 7, 22, 17, 30)},
                 'counsels': 'therapeutic',
                 'foregoing': 7634,
                 'linking': {'fructifies': {'Rushdie': False,
                                            'elided': 559,
                                            'endorsing': False,
                                            'nosediving': datetime.datetime(1987, 2, 2, 8, 31),
                                            'pivot': 'tarrier'},
                             'hairbreadth': datetime.datetime(1995, 1, 23, 9, 59),
                             'minimizing': {'Colosseum': True,
                                            'compacts': 'revolves',
                                            'piping': True,
                                            'repress': [],
                                            'violets': 'disassemble'},
                             'monaural': [3823, []],
                             'roiled': 6575},
                 'occasionally': ['balkier',
                                  datetime.datetime(1975, 7, 22, 4, 44),
                                  {'granulation': {},
                                   'groveling': 5952,
                                   'liquidized': True,
                                   'palsying': 'hems',
                                   'skid': datetime.datetime(2011, 11, 1, 16, 26)},
                                  [False, {}, 9829],
                                  True]}],
 'subscribed': datetime.datetime(1923, 6, 2, 15, 5),
 'success': {'Patricia': 2829,
             'conquistador': 'Ziegfeld',
             'indulges': True,
             'redrafts': {'ascertain': datetime.datetime(1967, 3, 1, 13, 42),
                          'divesting': {'alto': datetime.datetime(1949, 6, 25, 23, 50),
                                        'roadhouse': {'Bishop': True,
                                                      'allusive': datetime.datetime(1912, 1, 9, 15, 4),
                                                      'battened': 'overbooking',
                                                      'bloodthirstier': datetime.datetime(1995, 1, 23, 9, 59),
                                                      'decree': datetime.datetime(1923, 6, 2, 15, 5)},
                                        'sixties': 1285,
                                        'spool': None,
                                        'tinseling': True},
                          'instigated': 'sombrero',
                          'overspecialize': datetime.datetime(1983, 6, 27, 3, 56),
                          'seeds': 7127},
             'surfing': [8139, datetime.datetime(1930, 12, 12, 3, 58), 928]},
 'summered': True,
 'surrendering': {'crudit\xc3\xa9s': True,
                  'decrease': {'Cordoba': False,
                               'Plataea': False,
                               'Zairian': 'haves',
                               'debarks': datetime.datetime(1912, 1, 9, 15, 4),
                               'knocked': [4569, 5523, True, 8303, 5523]},
                  'prohibitory': {'Mach': True,
                                  'leakage': [['bone',
                                               datetime.datetime(1966, 7, 22, 17, 30),
                                               [],
                                               datetime.datetime(2011, 11, 1, 16, 26),
                                               {}],
                                              True,
                                              {'Fujitsu': False,
                                               'Morton': [],
                                               'bibulous': 'unjustifiable',
                                               'fractures': [],
                                               'grisly': {}},
                                              datetime.datetime(1911, 1, 17, 4, 24)],
                                  'mistime': 'daintiness',
                                  'probed': [],
                                  'urbaner': [{'Gucci': False,
                                               'Murat': 6605,
                                               'commerce': [],
                                               'growling': 3691,
                                               'nubile': 'maidenly'},
                                              {'Wassermann': False,
                                               'blackberry': 'publishes',
                                               'dissoluteness': True,
                                               'limier': 'drummer',
                                               'ponies': []},
                                              datetime.datetime(1949, 6, 25, 23, 50),
                                              ['nationwide', 'nets'],
                                              'dated']},
                  'waistline': 4120,
                  'zinced': 'dimwit'},
 'tanagers': None,
 'taxes': {'historically': [datetime.datetime(2013, 1, 22, 18, 59),
                            1129,
                            {'Grady': True,
                             'Mabel': 2253,
                             'adversary': datetime.datetime(1967, 3, 1, 13, 42),
                             'bares': False,
                             'caroms': 'entrepreneurial'},
                            {'disarm': {'Giza': [],
                                        'Mona': True,
                                        'Rapunzel': 'weltering',
                                        'radiologist': [],
                                        'raspberry': False},
                             'northerlies': {'Suleiman': 2825,
                                             'chicory': {},
                                             'feverishly': datetime.datetime(1995, 1, 23, 9, 59),
                                             'holsters': {},
                                             'threw': datetime.datetime(1987, 2, 2, 8, 31)},
                             'scale': [],
                             'stanch': 839,
                             'tune': 'evangelizing'},
                            True],
           'levering': datetime.datetime(1914, 5, 17, 9, 37),
           'teller': [{'Staples': 'danced',
                       'cocoon': 4437,
                       'secessionist': 6416,
                       'sways': 6023,
                       'venal': True},
                      [{'ambles': [],
                        'fetiches': 2096,
                        'secularize': {},
                        'sm\xc3\xb6rg\xc3\xa5sbord': datetime.datetime(1983, 6, 27, 3, 56),
                        'tartans': 34},
                       datetime.datetime(1911, 1, 17, 4, 24),
                       'sizzle',
                       8784]],
           'turret': 'moppets',
           'vacates': datetime.datetime(2011, 11, 1, 16, 26)},
 'teachable': 6100,
 'temporarily': None,
 'thoughtlessness': 2644,
 'totalled': 'extremer',
 'towheads': 5112,
 'traipsed': {'freeze': ['miscarried',
                         [['adjudication', [], 5849],
                          [datetime.datetime(1983, 6, 27, 3, 56), True],
                          6103,
                          [[], {}, None, True]],
                         False,
                         datetime.datetime(1918, 3, 3, 8, 35)],
              'hoses': False,
              'insinuate': 'handiest',
              'rethinks': 'fashion',
              'vibe': 'phished'},
 'twentieths': [],
 'unappetizing': datetime.datetime(1983, 6, 27, 3, 56),
 'unappreciated': True,
 'undernourished': [False,
                    [{'Diaspora': {'equivocating': 4363,
                                   'fingerings': 'Hannibal',
                                   'sorter': 'beatified',
                                   'tony': {},
                                   'variegate': True},
                      'Kurtis': 'adjudicated',
                      'chunks': {'approximately': 'winded',
                                 'brush': datetime.datetime(1918, 3, 3, 8, 35),
                                 'hack': datetime.datetime(1911, 1, 17, 4, 24),
                                 'spectacularly': datetime.datetime(2005, 7, 13, 17, 18),
                                 'tripe': 'untimelier'},
                      'effecting': 'bullfighters',
                      'loafer': {'Kentucky': datetime.datetime(1995, 1, 23, 9, 59),
                                 'Lenten': 462,
                                 'Tarkenton': [],
                                 'disappoints': {},
                                 'inked': 6087}}],
                    'wayfarer',
                    datetime.datetime(1949, 6, 25, 23, 50),
                    'goofing'],
 'unhealthful': 'toddling',
 'uninstaller': 'leach',
 'unquestionably': 'reprocesses',
 'unwieldiness': {'betrayer': [[{'jellybean': datetime.datetime(1966, 7, 22, 17, 30),
                                 'macroscopic': [],
                                 'redraws': {},
                                 'shapelessness': datetime.datetime(1966, 7, 22, 17, 30),
                                 'sting': {}},
                                True,
                                {'alley': {},
                                 'grouchy': True,
                                 'knees': {},
                                 'rakishly': 4758,
                                 'swashbuckling': 'chambers'},
                                datetime.datetime(1967, 3, 1, 13, 42),
                                True]],
                  'granules': ['litigates'],
                  'premiss': datetime.datetime(1995, 1, 23, 9, 59),
                  'tamed': {'Burgundian': 7837,
                            'Oceania': 'pedants',
                            'disuses': [True,
                                        9904,
                                        8025,
                                        {'exhilarate': True,
                                         'giveaways': [],
                                         'kneels': [],
                                         'poor': 'nationalists',
                                         'raglan': []}],
                            'packet': datetime.datetime(1912, 1, 9, 15, 4),
                            'verdict': datetime.datetime(2011, 11, 1, 16, 26)},
                  'therapeutic': 8212},
 'utilizes': {'Mitterrand': {'inclosing': datetime.datetime(1983, 6, 27, 3, 56),
                             'irately': datetime.datetime(1918, 3, 3, 8, 35),
                             'outsized': [{'Rx': {},
                                           'bare': 5115,
                                           'goats': 'Karakorum',
                                           'lessee': 'underlay',
                                           'typesets': 'judiciously'},
                                          False,
                                          {'concealment': 'distorting',
                                           'gracious': None,
                                           'host': [],
                                           'housemothers': True,
                                           'lamer': datetime.datetime(2011, 11, 1, 16, 26)}],
                             'rearmed': False,
                             'tallow': datetime.datetime(2013, 1, 22, 18, 59)},
              'hyphened': datetime.datetime(1911, 1, 17, 4, 24),
              'poverty': {'buoyancy': 7807,
                          'colorless': datetime.datetime(2005, 7, 13, 17, 18),
                          'finch': {'Meyers': {'anchorpersons': [],
                                               'breathe': [],
                                               'computational': True,
                                               'differentiation': datetime.datetime(1936, 12, 16, 10, 38),
                                               'warehouse': datetime.datetime(1966, 7, 22, 17, 30)},
                                    'arrives': False,
                                    'mausoleums': datetime.datetime(1966, 7, 22, 17, 30),
                                    'ponderous': 5684,
                                    'transmitted': True},
                          'mandarin': datetime.datetime(1930, 12, 12, 3, 58),
                          'touchy': datetime.datetime(1967, 3, 1, 13, 42)},
              'resurgences': datetime.datetime(1949, 6, 25, 23, 50),
              'whir': 'reenters'},
 'vichyssoise': [],
 'virus': {'Medicaids': 'gargle',
           'Pitt': 8044,
           'cormorant': 4672,
           'ministries': ['coffeecakes', True],
           'soy': 8861},
 'warred': datetime.datetime(1918, 3, 3, 8, 35),
 'weatherized': {'resilient': datetime.datetime(1966, 7, 22, 17, 30),
                 'snuggles': True,
                 'still': 'celesta',
                 'thefts': True,
                 'upliftings': {'Creator': datetime.datetime(2011, 11, 1, 16, 26),
                                'chevron': datetime.datetime(1930, 12, 12, 3, 58),
                                'mistranslated': datetime.datetime(1936, 12, 16, 10, 38),
                                'nationalized': datetime.datetime(1967, 3, 1, 13, 42),
                                'polygonal': [None, None, False, []]}},
 'whereof': 5169,
 'worry': 'rumples'}



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Zope Dublin Core PJ Backend Storage"""
from UserDict import DictMixin

import zope.interface
from zope.location import Location
from zope.dublincore.interfaces import IWriteZopeDublinCore
from zope.dublincore.zopedublincore import ZopeDublinCore
from zope.security.proxy import removeSecurityProxy

class DCDataWrapper(DictMixin):

    def __init__(self, data):
        self.data = data

    def __getitem__(self, key):
        return self.data[key.replace('.', '_')]

    def __setitem__(self, key, value):
        self.data[key.replace('.', '_')] = value

    def __delitem__(self, key):
        del self.data[key.replace('.', '_')]

    def keys(self):
        return [k.replace('_', '.') for k in self.data.keys()]


@zope.interface.implementer(IWriteZopeDublinCore)
class ZDCAnnotatableAdapter(ZopeDublinCore, Location):
    """Adapt annotatable objects to Zope Dublin Core."""
    DCKEY = 'dc'

    def __init__(self, context):
        self.__parent__ = context
        self.__name__ = self.DCKEY
        naked = removeSecurityProxy(context)
        if not hasattr(naked, self.__name__):
            setattr(naked, self.__name__, {})
        dcdata = DCDataWrapper(getattr(naked, self.__name__))
        super(ZDCAnnotatableAdapter, self).__init__(dcdata)

    def _changed(self):
        self.__parent__._p_changed = True



"""
taken from: http://justinfx.com/2012/07/25/python-2-7-3-serializer-speed-comparisons/
"""

import json
import pprint
from timeit import timeit

from pjpersist.tests import random_data
from pjpersist import testing
from pjpersist import serialize


DATA_ORIG = {
    'words': """
        Lorem ipsum dolor sit amet, consectetur adipiscing
        elit. Mauris adipiscing adipiscing placerat.
        Vestibulum augue augue,
        pellentesque quis sollicitudin id, adipiscing.
        """,
    'list': range(100),
    'dict': dict((str(i), 'a') for i in xrange(100)),
    'int': 100,
    'float': 100.123456
}

testing.setUpSerializers(None)
DATA_BIGDICT = serialize.ObjectWriter(None).get_state(random_data.BIGDICT)
DATA_HUGEDICT = serialize.ObjectWriter(None).get_state(random_data.HUGEDICT)

DATA = DATA_HUGEDICT

LOOPS = 1000

setup = "d=%r" % DATA

setup_pickle    = '%s ; import cPickle ; src = cPickle.dumps(d)' % setup
setup_pickle2   = '%s ; import cPickle ; src = cPickle.dumps(d, 2)' % setup
setup_json      = '%s ; import json; src = json.dumps(d)' % setup
setup_msgpack   = '%s ; src = msgpack.dumps(d)' % setup

tests = [
    # (title, setup, enc_test, dec_test)
    ('pickle (ascii)', 'import pickle; %s' % setup_pickle, 'pickle.dumps(d, 0)', 'pickle.loads(src)'),
    ('pickle (binary)', 'import pickle; %s' % setup_pickle2, 'pickle.dumps(d, 2)', 'pickle.loads(src)'),
    ('cPickle (ascii)', 'import cPickle; %s' % setup_pickle, 'cPickle.dumps(d, 0)', 'cPickle.loads(src)'),
    ('cPickle (binary)', 'import cPickle; %s' % setup_pickle2, 'cPickle.dumps(d, 2)', 'cPickle.loads(src)'),
    ('json', 'import json; %s' % setup_json, 'json.dumps(d)', 'json.loads(src)'),
    ('simplejson', 'import simplejson; %s' % setup_json, 'simplejson.dumps(d)', 'simplejson.loads(src)'),
    #('python-cjson-1.0.5', 'import cjson; %s' % setup_json, 'cjson.encode(d)', 'cjson.decode(src)'),
    ('ujson', 'import ujson; %s' % setup_json, 'ujson.dumps(d)', 'ujson.loads(src)'),
    #('yajl 0.3.5', 'import yajl; %s' % setup_json, 'yajl.dumps(d)', 'yajl.loads(src)'),
    #('msgpack-python-0.3.0', 'import msgpack; %s' % setup_msgpack, 'msgpack.dumps(d)', 'msgpack.loads(src)'),
]

def main():
    enc_table = []
    dec_table = []

    print("Running tests (%d LOOPS each)" % LOOPS)
    print("Data repr length: %d" % len(setup))

    for title, mod, enc, dec in tests:
        print(title)

        print("  [Encode]", enc)
        result = timeit(enc, mod, number=LOOPS)
        enc_table.append([title, result])

        print("  [Decode]", dec)
        result = timeit(dec, mod, number=LOOPS)
        dec_table.append([title, result])

    enc_table.sort(key=lambda x: x[1])
    enc_table.insert(0, ['Package', 'Seconds'])

    for x in enc_table:
        x[0] = x[0].ljust(20)

    dec_table.sort(key=lambda x: x[1])
    dec_table.insert(0, ['Package', 'Seconds'])

    for x in dec_table:
        x[0] = x[0].ljust(20)

    print("\nData repr length: %d" % len(setup))
    print("\nEncoding Test (%d LOOPS)" % LOOPS)
    pprint.pprint(enc_table)

    print("\nDecoding Test (%d LOOPS)" % LOOPS)
    pprint.pprint(dec_table)


results = """
results with DATA_ORIG:
------------------------
Encoding Test (15000 LOOPS)
[['Package             ', 'Seconds'],
 ['ujson               ', 0.15531587600708008],
 ['cPickle (binary)    ', 0.18461894989013672],
 ['json                ', 0.4842488765716553],
 ['simplejson          ', 0.6233742237091064],
 ['cPickle (ascii)     ', 0.9148178100585938],
 ['pickle (ascii)      ', 8.630627870559692],
 ['pickle (binary)     ', 8.879431009292603]]

Decoding Test (15000 LOOPS)
[['Package             ', 'Seconds'],
 ['cPickle (binary)    ', 0.23794102668762207],
 ['ujson               ', 0.24951386451721191],
 ['simplejson          ', 0.41387104988098145],
 ['cPickle (ascii)     ', 0.7791750431060791],
 ['json                ', 0.9328320026397705],
 ['pickle (binary)     ', 3.062810182571411],
 ['pickle (ascii)      ', 9.76430606842041]]

results with DATA_BIGDICT:
----------------------------
Encoding Test (15000 LOOPS)
[['Package             ', 'Seconds'],
 ['ujson               ', 0.18154096603393555],
 ['cPickle (binary)    ', 0.33770203590393066],
 ['json                ', 0.38971710205078125],
 ['simplejson          ', 0.6901121139526367],
 ['cPickle (ascii)     ', 0.928969144821167],
 ['pickle (ascii)      ', 6.54378604888916],
 ['pickle (binary)     ', 7.975723028182983]]

Decoding Test (15000 LOOPS)
[['Package             ', 'Seconds'],
 ['ujson               ', 0.23072004318237305],
 ['cPickle (binary)    ', 0.26100707054138184],
 ['simplejson          ', 0.2880890369415283],
 ['cPickle (ascii)     ', 0.4758889675140381],
 ['json                ', 0.891761064529419],
 ['pickle (binary)     ', 3.5709359645843506],
 ['pickle (ascii)      ', 6.168566942214966]]

results with DATA_HUGEDICT:
---------------------------
Data repr length: 37184

Encoding Test (1000 LOOPS)
[['Package             ', 'Seconds'],
 ['ujson               ', 0.28696608543395996],
 ['json                ', 0.49742913246154785],
 ['cPickle (binary)    ', 0.5122489929199219],
 ['simplejson          ', 0.9091768264770508],
 ['cPickle (ascii)     ', 1.2010438442230225],
 ['pickle (ascii)      ', 8.14183497428894],
 ['pickle (binary)     ', 10.854649066925049]]

Decoding Test (1000 LOOPS)
[['Package             ', 'Seconds'],
 ['ujson               ', 0.4056370258331299],
 ['cPickle (binary)    ', 0.41997385025024414],
 ['simplejson          ', 0.44395899772644043],
 ['cPickle (ascii)     ', 0.71999192237854],
 ['json                ', 1.1725149154663086],
 ['pickle (binary)     ', 4.6916890144348145],
 ['pickle (ascii)      ', 7.210031986236572]]

"""



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Python Serializers for common objects with weird reduce output."""
import datetime
from pjpersist import serialize


class DateSerializer(serialize.ObjectSerializer):

    fmt = "%Y-%m-%d"

    def can_read(self, state):
        return isinstance(state, dict) and \
               state.get('_py_type') == 'datetime.date'

    def read(self, state):
        return datetime.datetime.strptime(state['value'], self.fmt).date()

    def can_write(self, obj):
        return isinstance(obj, datetime.date)

    def write(self, obj):
        return {'_py_type': 'datetime.date',
                'value': obj.strftime(self.fmt)}


class TimeSerializer(serialize.ObjectSerializer):

    fmt = "%H:%M:%S"

    def can_read(self, state):
        return isinstance(state, dict) and \
               state.get('_py_type') == 'datetime.time'

    def read(self, state):
        return datetime.datetime.strptime(state['value'], self.fmt).time()

    def can_write(self, obj):
        return isinstance(obj, datetime.time)

    def write(self, obj):
        return {'_py_type': 'datetime.time',
                'value': obj.strftime(self.fmt)}


class DateTimeSerializer(serialize.ObjectSerializer):

    # XXX: timezone?
    fmt = "%Y-%m-%dT%H:%M:%S"

    def can_read(self, state):
        return isinstance(state, dict) and \
               state.get('_py_type') == 'datetime.datetime'

    def read(self, state):
        return datetime.datetime.strptime(state['value'], self.fmt)

    def can_write(self, obj):
        return isinstance(obj, datetime.datetime)

    def write(self, obj):
        return {'_py_type': 'datetime.datetime',
                'value': obj.strftime(self.fmt)}
